from __future__ import annotations

import abc
import asyncio
import dataclasses
import ipaddress
from typing import (
    Any,
    Awaitable,
    Dict,
    Iterable,
    Optional,
    Sequence,
    Set,
    Tuple,
    TypeVar,
    cast,
)

import aiobotocore.session
import boto3

from meadowrun.aws_integration.aws_core import (
    MeadowrunAWSAccessError,
    MeadowrunNotInstalledError,
    _get_current_ip_for_ssh,
    _get_default_region_name,
)
from meadowrun.aws_integration.ec2_pricing import _get_ec2_instance_types
from meadowrun.aws_integration.management_lambdas.ec2_alloc_stub import (
    _EC2_ALLOC_TAG,
    _EC2_ALLOC_TAG_VALUE,
    ignore_boto3_error_code,
    ignore_boto3_error_code_async,
)
from meadowrun.aws_integration.quotas import (
    get_quota_for_instance_type,
    instance_family_from_type,
)
from meadowrun.instance_selection import (
    CloudInstance,
    CloudInstanceType,
    OnDemandOrSpotType,
    ResourcesInternal,
    choose_instance_types_for_job,
)
from meadowrun.version import __version__


_T = TypeVar("_T")


# A security group that allows SSH, clients' IP addresses get added as needed
_MEADOWRUN_SSH_SECURITY_GROUP = "meadowrun_ssh_security_group"


def get_ssh_security_group_id(region_name: str, vpc_id: Optional[str] = None) -> str:
    """
    Gets the id of the meadowrun SSH security group, which determines which IPs are
    allowed to SSH into the Meadowrun-managed instances.
    """
    ec2_resource = boto3.resource("ec2", region_name=region_name)
    security_group = _get_ec2_security_group(
        ec2_resource, _MEADOWRUN_SSH_SECURITY_GROUP, vpc_id
    )
    if security_group is None:
        raise MeadowrunNotInstalledError(
            f"security group {_MEADOWRUN_SSH_SECURITY_GROUP}"
        )
    return security_group.id


def _authorize_current_ip_for_security_group(
    security_group: Any, from_port: int, to_port: int, current_ip: str
) -> None:
    # check if our ip is already in the security group. This isn't watertight, as we
    # might interpret these rules differently from how AWS does, and there are other
    # options for granting access, e.g. via a security group. So we should usually catch
    # any exceptions in the caller
    already_authorized = False
    current_ip_for_ssh_address = ipaddress.ip_address(current_ip)
    for permission in security_group.ip_permissions:
        if permission["FromPort"] <= from_port and to_port <= permission["ToPort"]:
            for ip_range in permission.get("IpRanges", ()):
                cidr_ip = ip_range.get("CidrIp")
                if cidr_ip and current_ip_for_ssh_address in ipaddress.ip_network(
                    cidr_ip
                ):
                    already_authorized = True
                    break
        if already_authorized:
            break

    if not already_authorized:
        ignore_boto3_error_code(
            lambda: security_group.authorize_ingress(
                IpProtocol="tcp",
                CidrIp=f"{current_ip}/32",
                FromPort=from_port,
                ToPort=to_port,
            ),
            "InvalidPermission.Duplicate",
        )
        print(
            f"Authorized the current ip address {current_ip} for ports "
            f"{from_port}-{to_port} to the {security_group.group_name} security group"
        )


async def authorize_current_ip_for_meadowrun_ssh(
    region_name: str, subnet_id: Optional[str] = None
) -> None:
    """
    Tries to add the current IP address to the list of IPs allowed to SSH into the
    meadowrun SSH security group (will warn rather than raise on error). Returns the
    security group id of the meadowrun SSH security group.
    """

    current_ip = "<your ip address>"
    try:
        ec2_resource = boto3.resource("ec2", region_name=region_name)

        if subnet_id is not None:
            # TODO this code is duplicated in launch_instances
            vpc_id = ec2_resource.Subnet(subnet_id).vpc_id
        else:
            vpc_id = None

        security_group = _get_ec2_security_group(
            ec2_resource, _MEADOWRUN_SSH_SECURITY_GROUP, vpc_id
        )
        if security_group is None:
            raise MeadowrunNotInstalledError(
                f"security group {_MEADOWRUN_SSH_SECURITY_GROUP}"
            )

        current_ip = await _get_current_ip_for_ssh()
        _authorize_current_ip_for_security_group(security_group, 22, 22, current_ip)
    except asyncio.CancelledError:
        raise
    except Exception as e:
        print(
            "Warning, failed to authorize current IP address for SSH. Connecting to"
            " Meadowrun instances will fail unless your connection has been "
            "authorized in a different way. Most likely, meadowrun was installed "
            "with `meadowrun-manage-ec2 install --allow-authorize-ips False`. If "
            "this is the case, you can rerun with `--allow-authorize-ips True`, or "
            "have an administrator manually authorize your connection to the "
            f"{_MEADOWRUN_SSH_SECURITY_GROUP} security group by running something "
            f"like: `aws ec2 authorize-security-group-ingress --group-name "
            f"{_MEADOWRUN_SSH_SECURITY_GROUP} --protocol tcp --port 22 --cidr "
            f"{current_ip}/32` {e}"
        )


def ensure_security_group(
    group_name: str, region_name: str, vpc_id: Optional[str] = None
) -> str:
    """
    Creates the specified security group if it doesn't exist. If it does exist, does not
    modify it (as there may be existing rules that should not be deleted).

    Returns the id of the security group.
    """
    ec2_resource = boto3.resource("ec2", region_name=region_name)
    security_group = _get_ec2_security_group(ec2_resource, group_name, vpc_id)
    if security_group is None:
        if vpc_id is not None:
            additional_parameters = {"VpcId": vpc_id}
        else:
            additional_parameters = {}
        security_group = ec2_resource.create_security_group(
            Description=group_name,
            GroupName=group_name,
            TagSpecifications=[
                {
                    "ResourceType": "security-group",
                    "Tags": [{"Key": _EC2_ALLOC_TAG, "Value": _EC2_ALLOC_TAG_VALUE}],
                }
            ],
            **additional_parameters,
        )
    return security_group.id


async def ensure_port_security_group(ports: str, region_name: str) -> str:
    """
    Creates a security group for opening the specified ports to the current ip address
    as well as other machines in that security group. Returns a security group id.
    """
    group_name = f"meadowrun{ports}"

    from_port, sep, to_port = ports.partition("-")
    if sep == "-":
        from_port_int = int(from_port)
        to_port_int = int(to_port)
    else:
        from_port_int = int(ports)
        to_port_int = from_port_int

    ec2_resource = boto3.resource("ec2", region_name)
    security_group = _get_ec2_security_group(ec2_resource, group_name)
    if security_group is None:
        try:
            security_group = ec2_resource.create_security_group(
                Description=group_name,
                GroupName=group_name,
                TagSpecifications=[
                    {
                        "ResourceType": "security-group",
                        "Tags": [
                            {"Key": _EC2_ALLOC_TAG, "Value": _EC2_ALLOC_TAG_VALUE}
                        ],
                    }
                ],
            )
        except Exception as e:
            raise MeadowrunAWSAccessError(
                "Unable to create a security group. Please ask an administrator to run "
                "`meadowrun-manage-ec2 install --allow-authorize-ips` or alternatively,"
                " ask them to manually create this security group and authorize your "
                "access to it with the following commands: `aws ec2 "
                f"create-security-group --description {group_name} --group-name "
                f"{group_name}` then `aws ec2 authorize-security-group-ingress "
                f"--group-name {group_name} --protocol tcp --port {ports} "
                f"--source-group {group_name}` then finally `aws ec2 "
                f"authorize-security-group-ingress --group-name {group_name} --protocol"
                f" tcp --port {ports} --cidr {await _get_current_ip_for_ssh()}/32`",
                True,
            ) from e

        ignore_boto3_error_code(
            lambda: security_group.authorize_ingress(
                IpPermissions=[
                    {
                        "FromPort": from_port_int,
                        "ToPort": to_port_int,
                        "IpProtocol": "tcp",
                        "UserIdGroupPairs": [{"GroupId": security_group.id}],
                    }
                ]
            ),
            "InvalidPermission.Duplicate",
        )

    current_ip = await _get_current_ip_for_ssh()
    try:
        _authorize_current_ip_for_security_group(
            security_group, from_port_int, to_port_int, current_ip
        )
    except Exception as e:
        print(
            f"Warning, failed to authorize current IP address for ports {ports}. "
            "Most likely you will fail to connect to your services on these ports. "
            "Please ask an administrator to run `meadowrun-manage-ec2 install "
            "--allow-authorize-ips` or alternatively, ask them to manually authorize "
            "your access by running: `aws ec2 authorize-security-group-ingress "
            f"--group-name {group_name} --protocol tcp --port {ports} --cidr "
            f"{current_ip}/32` {e}",
        )

    return security_group.id


class LaunchEC2InstanceResult(abc.ABC):
    pass


@dataclasses.dataclass(frozen=True)
class LaunchEC2InstanceSuccess(LaunchEC2InstanceResult):
    """
    Indicates that launch_ec2_instance succeeded. public_address_continuation can be
    awaited to get the public address of the launched instance.
    public_address_continuation can be None which indicates that the launch of the
    instance was cancelled. If the instance was mid-launch, then the instance was
    terminated before it could fully launch.
    """

    public_address_continuation: Awaitable[Optional[str]]
    instance_id: str


@dataclasses.dataclass(frozen=True)
class LaunchEC2InstanceAborted(LaunchEC2InstanceResult):
    """Indicates that launch_ec2_instance was aborted"""

    pass


class LaunchEC2InstanceTypeUnavailable(LaunchEC2InstanceResult):
    """
    Indicates that launch_ec2_instance failed because the specified instance type is not
    available.
    """

    @abc.abstractmethod
    def get_message(self) -> str:
        """Explains the failure"""
        pass

    @abc.abstractmethod
    def unusable_instance_types(
        self, all_instance_types: Iterable[CloudInstanceType]
    ) -> Iterable[Tuple[str, OnDemandOrSpotType]]:
        """
        Given all instance types, returns the instance types that are now considered
        unusable because we encountered this error. E.g. for a quota issue, we should
        ignore any instance types that are under the same quota that is being hit.
        """
        pass


@dataclasses.dataclass(frozen=True)
class LaunchEC2InstanceAwsCapacity(LaunchEC2InstanceTypeUnavailable):
    """
    Indicates that an EC2 instance couldn't be launched because of AWS capacity
    constraints
    """

    message: str
    instance_type: str
    on_demand_or_spot: OnDemandOrSpotType

    def get_message(self) -> str:
        return self.message

    def unusable_instance_types(
        self, all_instance_types: Iterable[CloudInstanceType]
    ) -> Iterable[Tuple[str, OnDemandOrSpotType]]:
        yield self.instance_type, self.on_demand_or_spot


@dataclasses.dataclass(frozen=True)
class LaunchEC2InstanceQuota(LaunchEC2InstanceTypeUnavailable):
    """Indicates that an EC2 instance couldn't be launched because a quota was hit"""

    message: str
    quota_code: str
    quota_instance_families: Tuple[str, ...]

    def get_message(self) -> str:
        return self.message

    def unusable_instance_types(
        self, all_instance_types: Iterable[CloudInstanceType]
    ) -> Iterable[Tuple[str, OnDemandOrSpotType]]:
        return (
            (instance_type.name, instance_type.on_demand_or_spot)
            for instance_type in all_instance_types
            if instance_family_from_type(instance_type.name)
            in self.quota_instance_families
        )


@dataclasses.dataclass(frozen=True)
class LaunchEC2InstanceAmiNotAvailable(LaunchEC2InstanceResult):
    ami_id: str


@dataclasses.dataclass(frozen=True)
class LaunchEC2InstanceSettings:
    """
    This class contains the settings that come from AllocEC2Instance. See that class for
    details on the semantics of each argument.
    """

    ami_id: str
    subnet_id: Optional[str] = None
    security_group_ids: Optional[Sequence[str]] = None
    iam_role_instance_profile: Optional[str] = None


async def launch_ec2_instance(
    region_name: str,
    instance_type: str,
    on_demand_or_spot: OnDemandOrSpotType,
    launch_settings: LaunchEC2InstanceSettings,
    user_data: Optional[str] = None,
    key_name: Optional[str] = None,
    tags: Optional[Dict[str, str]] = None,
    volume_size_gb: int = 100,
    abort: Optional[asyncio.Event] = None,
) -> LaunchEC2InstanceResult:
    """
    Launches the specified EC2 instance. See derived classes of LaunchEC2InstanceResult
    for possible return values.
    """

    optional_args: Dict[str, Any] = {
        # TODO allow users to specify the size of the EBS they need
        "BlockDeviceMappings": [
            {
                "DeviceName": "/dev/sda1",
                "Ebs": {
                    "DeleteOnTermination": True,
                    "VolumeSize": volume_size_gb,
                    "VolumeType": "gp2",
                },
            }
        ]
    }
    if launch_settings.subnet_id:
        # this should not be set if we specify a custom network interface (which isn't
        # supported yet)
        optional_args["SubnetId"] = launch_settings.subnet_id
    if launch_settings.security_group_ids:
        optional_args["SecurityGroupIds"] = launch_settings.security_group_ids
    if launch_settings.iam_role_instance_profile:
        optional_args["IamInstanceProfile"] = {
            "Name": launch_settings.iam_role_instance_profile
        }
    if key_name:
        optional_args["KeyName"] = key_name
    if user_data:
        optional_args["UserData"] = user_data
    if tags:
        optional_args["TagSpecifications"] = [
            {
                "ResourceType": "instance",
                "Tags": [{"Key": key, "Value": value} for key, value in tags.items()],
            }
        ]

    if on_demand_or_spot == "on_demand":
        pass
    elif on_demand_or_spot == "spot":
        optional_args["InstanceMarketOptions"] = {"MarketType": "spot"}
    else:
        raise ValueError(f"Unexpected value for on_demand_or_spot {on_demand_or_spot}")

    if abort is not None and abort.is_set():
        return LaunchEC2InstanceAborted()

    async with aiobotocore.session.get_session().create_client(
        "ec2", region_name=region_name
    ) as ec2_client:
        # https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/ec2.html#EC2.Client.run_instances
        success, instances, error_code = await ignore_boto3_error_code_async(
            ec2_client.run_instances(
                ImageId=launch_settings.ami_id,
                MinCount=1,
                MaxCount=1,
                InstanceType=instance_type,  # type: ignore
                **optional_args,
            ),
            {
                "InsufficientInstanceCapacity",
                "MaxSpotInstanceCountExceeded",
                "VcpuLimitExceeded",
                "InvalidAMIID.NotFound",
            },
        )
        if not success:
            if error_code == "InsufficientInstanceCapacity":
                return LaunchEC2InstanceAwsCapacity(
                    f"AWS does not have enough capacity in region {region_name} to "
                    f"create a {on_demand_or_spot} instance of type {instance_type}",
                    instance_type,
                    on_demand_or_spot,
                )
            elif error_code in ("MaxSpotInstanceCountExceeded", "VcpuLimitExceeded"):
                # I believe these error codes have the same meaning for spot/on-demand
                # instances
                return LaunchEC2InstanceQuota(
                    *get_quota_for_instance_type(
                        instance_type, on_demand_or_spot, region_name
                    )
                )
            elif error_code == "InvalidAMIID.NotFound":
                return LaunchEC2InstanceAmiNotAvailable(launch_settings.ami_id)
            else:
                raise ValueError(f"Unexpected boto3 error code {error_code}")

    # types_aiobotocore seems to have the wrong types for instances
    instances_untyped = cast(Any, instances)
    if len(instances_untyped["Instances"]) != 1:
        raise ValueError(
            "run_instances succeeded but returned "
            f"{len(instances_untyped['Instances'])} instances"
        )
    instance_id = instances_untyped["Instances"][0]["InstanceId"]
    return LaunchEC2InstanceSuccess(
        _launch_instance_continuation(instance_id, region_name, abort), instance_id
    )


async def describe_single_instance(ec2_client: Any, instance_id: str) -> Dict:
    response = await ec2_client.describe_instances(InstanceIds=[instance_id])
    reservations = response["Reservations"]
    if len(reservations) != 1:
        raise ValueError(f"Unexpected number of reservations: {len(reservations)}")
    instances = reservations[0]["Instances"]
    if len(instances) != 1:
        raise ValueError(f"Unexpected number of instances: {len(instances)}")
    return instances[0]


async def _launch_instance_continuation(
    instance_id: str, region_name: str, abort: Optional[asyncio.Event]
) -> Optional[str]:
    """
    instance should be a boto3 EC2 instance, waits for the instance to be running and
    then returns its public DNS name
    """

    # ideally we would reuse the ec2_client from the caller, but this is much simpler
    i = 0
    async with aiobotocore.session.get_session().create_client(
        "ec2", region_name=region_name
    ) as ec2_client:
        try:
            # this is a reimplementation of
            # https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/ec2.html#EC2.Instance.wait_until_running
            # main difference is that we check every 7 seconds rather than every 15
            # seconds
            while True:
                await asyncio.sleep(7)
                # ideally we would interrupt the above sleep if cancel becomes set
                if abort is not None and abort.is_set():
                    await ec2_client.terminate_instances(InstanceIds=[instance_id])
                    return None

                instance = await describe_single_instance(ec2_client, instance_id)
                if instance["State"]["Name"] == "running":
                    break
                elif instance["State"]["Name"] == "pending":
                    pass
                else:
                    print(
                        f"Unexpected instance state while waiting for {instance_id} to "
                        f"start up: {instance['State']}. This instance may never start "
                        "up."
                    )

                i += 1
                if i >= 60:
                    raise TimeoutError(
                        "Checked 60 times, waiting 7 seconds between each check and "
                        f"instance {instance_id} is still not running"
                    )

            if instance["PublicDnsName"]:
                dns_name = instance["PublicDnsName"]
            else:
                # try one more time after waiting 1 seconds
                await asyncio.sleep(1.0)
                instance = await describe_single_instance(ec2_client, instance_id)
                if instance["PublicDnsName"]:
                    dns_name = instance["PublicDnsName"]
                elif instance["PrivateDnsName"]:
                    print(
                        "Warning, no public DNS name available for instance "
                        f"{instance_id}. Will try to use the private DNS name, but this"
                        " will only work if the current machine is in the same VPC as "
                        "the new instance. If this message is unexpected, please try "
                        "setting your subnet to automatically assign public IPv4 "
                        "addresses"
                    )
                    dns_name = instance["PrivateDnsName"]
                else:
                    raise ValueError(
                        f"Instance {instance_id} is running, but does not have a public"
                        " DNS name or a private DNS name."
                    )
        except BaseException:
            # This is a best effort at terminating the instance. If we don't manage to
            # terminate, it will get cleaned up by the management lambda as the instance
            # will exist but not be registered. There's code before and after this that
            # we should also wrap in a similar try/except block, but this is by far the
            # most likely place where there will be an issue.
            await ec2_client.terminate_instances(InstanceIds=[instance_id])
            raise
    return dns_name


async def launch_ec2_instances(
    instance_type_resources_required_per_job: ResourcesInternal,
    num_jobs: int,
    launch_settings: LaunchEC2InstanceSettings,
    region_name: Optional[str] = None,
    user_data: Optional[str] = None,
    key_name: Optional[str] = None,
    tags: Optional[Dict[str, str]] = None,
    abort: Optional[asyncio.Event] = None,
) -> Sequence[CloudInstance]:
    """
    Launches enough EC2 instances to run num_jobs jobs that each require the specified
    amount of CPU/memory. Returns a sequence of CloudInstance.
    CloudInstance.instance_type.workers_per_instance_full indicates the maximum number
    of jobs that can run on that instance. The sum of all of the
    workers_per_instance_full will be greater than or equal to the original num_jobs
    parameter.
    """

    if region_name is None:
        region_name = await _get_default_region_name()

    instance_types = await _get_ec2_instance_types(region_name)

    num_jobs_left_to_allocate = num_jobs
    public_dns_name_tasks = []
    instance_ids = []
    launched_instance_types_repeated = []

    # As we try to launch instances, we will realize that some of them cannot be
    # launched due to lack of capacity in AWS or quotas. We'll keep track of what
    # instance types aren't actually available in this variable
    unusable_instance_types: Set[Tuple[str, OnDemandOrSpotType]] = set()

    while num_jobs_left_to_allocate > 0:
        if unusable_instance_types:
            print(
                "Trying to proceed with less optimal instance types. This can result in"
                " significantly more expensive instances"
            )

        chosen_instance_types = choose_instance_types_for_job(
            instance_type_resources_required_per_job,
            num_jobs_left_to_allocate,
            (
                i
                for i in instance_types
                if (i.name, i.on_demand_or_spot) not in unusable_instance_types
            ),
        )

        at_least_one_chosen_instance_type = False
        num_jobs_left_to_allocate = 0
        for instance_type in chosen_instance_types:
            if abort is not None and abort.is_set():
                break

            at_least_one_chosen_instance_type = True
            for _ in range(instance_type.num_instances):
                launch_ec2_result = await launch_ec2_instance(
                    region_name,
                    instance_type.instance_type.name,
                    instance_type.instance_type.on_demand_or_spot,
                    launch_settings,
                    user_data=user_data,
                    key_name=key_name,
                    tags=tags,
                    abort=abort,
                )
                if isinstance(launch_ec2_result, LaunchEC2InstanceSuccess):
                    public_dns_name_tasks.append(
                        launch_ec2_result.public_address_continuation
                    )
                    instance_ids.append(launch_ec2_result.instance_id)
                    launched_instance_types_repeated.append(instance_type)
                elif isinstance(launch_ec2_result, LaunchEC2InstanceAborted):
                    break
                elif isinstance(launch_ec2_result, LaunchEC2InstanceTypeUnavailable):
                    print(launch_ec2_result.get_message())

                    # TODO this isn't exactly right, we can end up over-allocating
                    # because we might choose a larger instance than we "actually"
                    # needed, but this shouldn't be a huge factor
                    num_jobs_left_to_allocate += instance_type.workers_per_instance_full
                    unusable_instance_types.update(
                        launch_ec2_result.unusable_instance_types(instance_types)
                    )
                elif isinstance(launch_ec2_result, LaunchEC2InstanceAmiNotAvailable):
                    raise ValueError(
                        "Launching an EC2 instance failed because the AMI ID "
                        f"{launch_ec2_result.ami_id} was not found. You may need to "
                        "update Meadowrun, as we regularly remove the AMIs for old "
                        "versions. Please contact us (contact@meadowdata.io) if you "
                        "need support for old versions. You are on version "
                        f"{__version__}. (This message does not apply if you are "
                        "providing a custom ami_id.)"
                    )
                else:
                    raise ValueError(
                        "Unexpected type of launch_ec2_result "
                        f"{type(launch_ec2_result)}"
                    )

        if not at_least_one_chosen_instance_type:
            raise ValueError(
                "There were no instance types that could be selected for: "
                f"{instance_type_resources_required_per_job}"
            )

    public_dns_names = await asyncio.gather(*public_dns_name_tasks)

    return [
        CloudInstance(public_dns_name, instance_id, instance_type)
        for public_dns_name, instance_id, instance_type in zip(
            public_dns_names, instance_ids, launched_instance_types_repeated
        )
        # None indicates that the instance launch was cancelled
        if public_dns_name is not None
    ]


def _get_ec2_security_group(
    ec2_resource: Any, name: str, vpc_id: Optional[str] = None
) -> Any:
    """
    Gets the specified security group if it exists. Returns an
    Optional[boto3.resources.factory.ec2.SecurityGroup] (not in the type signature
    because boto3 uses dynamic types).
    """
    if vpc_id is None:
        # check the default VPC
        success, matching_groups = ignore_boto3_error_code(
            lambda: list(ec2_resource.security_groups.filter(GroupNames=[name])),
            "InvalidGroup.NotFound",
        )
        if not success:
            return None
    else:
        all_groups = list(
            ec2_resource.security_groups.filter(
                Filters=[{"Name": "vpc-id", "Values": [vpc_id]}]
            )
        )
        matching_groups = [group for group in all_groups if group.group_name == name]

    assert matching_groups is not None  # just for mypy
    if len(matching_groups) == 0:
        return None
    elif len(matching_groups) > 1:
        raise ValueError(
            f"Found multiple security groups with the same name {name} which was "
            f"unexpected"
        )
    else:
        return matching_groups[0]



from __future__ import annotations

from typing import Optional, TYPE_CHECKING

from meadowrun.aws_integration.ec2_instance_allocation import AllocEC2Instance
from meadowrun.azure_integration.azure_instance_allocation import AllocAzureVM

if TYPE_CHECKING:
    from meadowrun.run_job_core import CloudProviderType, AllocVM


def AllocCloudInstance(
    cloud_provider: CloudProviderType, region_name: Optional[str] = None
) -> AllocVM:
    """
    This function is deprecated and will be removed in version 1.0.0. Please use
    [AllocEC2Instance][meadowrun.AllocEC2Instance] or
    [AllocAzureVM][meadowrun.AllocAzureVM] directly.
    """
    if cloud_provider == "EC2":
        return AllocEC2Instance(region_name)
    elif cloud_provider == "AzureVM":
        return AllocAzureVM(region_name)
    else:
        raise ValueError(f"Unexpected value for cloud_provider {cloud_provider}")



from __future__ import annotations

import asyncio
import pickle
import shutil
import sys
import traceback
import zipfile
from typing import Optional, Tuple, TypeVar, IO, TYPE_CHECKING


from meadowrun.meadowrun_pb2 import ProcessState

if TYPE_CHECKING:
    from os import PathLike
    from typing_extensions import Literal


def pickle_exception(e: Exception, pickle_protocol: int) -> bytes:
    """
    We generally don't want to pickle exceptions directly--there's no guarantee that a
    random exception that was thrown can be unpickled in a different process.
    """
    tb = "".join(traceback.format_exception(type(e), e, e.__traceback__))
    return pickle.dumps(
        (str(type(e)), str(e), tb),
        protocol=pickle_protocol,
    )


def unpickle_exception(bs: bytes) -> Tuple[str, str, str]:
    return pickle.loads(bs)


COMPLETED_PROCESS_STATES = {
    ProcessState.ProcessStateEnum.SUCCEEDED,
    ProcessState.ProcessStateEnum.RUN_REQUEST_FAILED,
    ProcessState.ProcessStateEnum.PYTHON_EXCEPTION,
    ProcessState.ProcessStateEnum.NON_ZERO_RETURN_CODE,
    ProcessState.ProcessStateEnum.RESOURCES_NOT_AVAILABLE,
    ProcessState.ProcessStateEnum.ERROR_GETTING_STATE,
}


_T = TypeVar("_T")


def assert_is_not_none(resources: Optional[_T]) -> _T:
    """A helper for mypy"""
    assert resources is not None
    return resources


def remove_corrupted_environment(path: str) -> None:
    try:
        shutil.rmtree(path, True)
    except asyncio.CancelledError:
        raise
    except BaseException:
        print(
            f"Warning, exception trying to delete {path}, this environment is corrupted"
        )
        traceback.print_exc()


async def cancel_task(task: asyncio.Task) -> None:
    """Cancel a task, wait for it, don't throw CancelledError."""
    try:
        task.cancel()
        await task
    except asyncio.CancelledError:
        pass


def create_zipfile(
    file: str | PathLike[str] | IO[bytes],
    mode: Literal["r", "w", "x", "a"],
    compression: int,
) -> zipfile.ZipFile:
    """
    Equivalent to zipfile.ZipFile(...) but passes strict_timestamps=False if that
    parameter is available in the current version of python
    """
    if sys.version_info < (3, 8):
        # strict_timestamps parameter was addded in 3.8:
        # https://docs.python.org/3/library/zipfile.html#zipfile.ZipFile
        return zipfile.ZipFile(file, mode, compression)
    else:
        return zipfile.ZipFile(file, mode, compression, strict_timestamps=False)



from __future__ import annotations

import asyncio
import os
import shutil
import sys
from typing import Callable, Awaitable

import filelock

from meadowrun.shared import remove_corrupted_environment
from meadowrun.storage_keys import STORAGE_ENV_CACHE_PREFIX
from meadowrun.deployment.prerequisites import (
    EnvironmentSpecPrerequisites,
    GOOGLE_AUTH_PACKAGE,
)

_POETRY_ENVIRONMENT_TIMEOUT = 10 * 60


async def get_cached_or_create_poetry_environment(
    environment_hash: str,
    project_file_path: str,
    prerequisites: EnvironmentSpecPrerequisites,
    new_environment_path: str,
    try_get_file: Callable[[str, str], Awaitable[bool]],
    upload_file: Callable[[str, str], Awaitable[None]],
) -> str:
    """
    If the desired poetry environment exists, does nothing. If the environment has been
    cached, creates it from the cache. Otherwise creates the environment from scratch
    and caches it. Returns the path to the newly created python interpreter.

    try_get_file and upload_file are for interacting with the cache, see
    compile_environment_spec_locally for more details
    """
    # this assumes that that current version of python is what will be used in
    # create_pip_environment
    new_environment_path = (
        f"{new_environment_path}-{sys.version_info.major}.{sys.version_info.minor}"
    )

    venv_path = os.path.join(new_environment_path, ".venv")
    new_environment_interpreter = os.path.join(venv_path, "bin", "python")

    with filelock.FileLock(f"{new_environment_path}.lock", _POETRY_ENVIRONMENT_TIMEOUT):
        if os.path.exists(new_environment_path):
            return new_environment_interpreter

        remote_cached_file_name = (
            f"{STORAGE_ENV_CACHE_PREFIX}{environment_hash}-{sys.version_info.major}."
            f"{sys.version_info.minor}.tar.gz"
        )
        local_cached_file = f"{new_environment_path}.tar.gz"
        download_succeeded = await try_get_file(
            remote_cached_file_name, local_cached_file
        )
        if download_succeeded:
            try:
                print("Unpacking cached poetry environment")
                os.makedirs(venv_path, exist_ok=True)
                try:
                    # TODO maybe cleaner to use the built-in python tar libraries?
                    return_code = await (
                        await asyncio.create_subprocess_exec(
                            "tar", "-xzf", local_cached_file, "-C", venv_path
                        )
                    ).wait()
                    if return_code != 0:
                        raise ValueError(
                            f"Unpacking cached pip environment {local_cached_file} "
                            f"returned code {return_code}"
                        )
                    return new_environment_interpreter
                except BaseException:
                    remove_corrupted_environment(new_environment_path)
                    raise
            finally:
                try:
                    os.remove(local_cached_file)
                except asyncio.CancelledError:
                    raise
                except BaseException:
                    pass

        print("Creating the poetry environment")
        try:
            await create_poetry_environment(
                project_file_path, new_environment_path, prerequisites
            )
        except BaseException:
            remove_corrupted_environment(new_environment_path)
            raise

        try:
            import venv_pack  # see note on reference in pyproject.toml
        except ImportError:
            print(
                "Warning unable to cache poetry environment because venv_pack is "
                "missing"
            )
            return new_environment_interpreter

        # TODO we shouldn't wait for this to start running the job but we also shouldn't
        # kill the container until this finishes
        print("Caching the poetry environment")
        try:
            # there might be an old file from a failed attempt
            os.remove(local_cached_file)
        except OSError:
            pass
        venv_pack.pack(venv_path, output=local_cached_file)
        await upload_file(local_cached_file, remote_cached_file_name)

        return new_environment_interpreter


# this path depends on the poetry installation script, we have to hope that this never
# changes
_POETRY_PATH = "/root/.local/bin/poetry"


async def create_poetry_environment(
    project_file_path: str,
    new_environment_path: str,
    prerequisites: EnvironmentSpecPrerequisites,
) -> None:
    os.makedirs(new_environment_path, exist_ok=True)

    for file in ("pyproject.toml", "poetry.lock"):
        shutil.copyfile(
            os.path.join(project_file_path, file),
            os.path.join(new_environment_path, file),
        )

    if prerequisites & EnvironmentSpecPrerequisites.GOOGLE_AUTH:
        return_code = await (
            await asyncio.create_subprocess_exec(
                _POETRY_PATH,
                "self",
                "add",
                GOOGLE_AUTH_PACKAGE,
            )
        ).wait()
        if return_code != 0:
            raise ValueError(
                "poetry environment prerequisites installation failed with return code "
                f"{return_code}"
            )

    # this code is roughly equivalent to the code in PoetryDockerfile and
    # PoetryDockerfile
    return_code = await (
        await asyncio.create_subprocess_exec(
            _POETRY_PATH, "install", "--no-root", cwd=new_environment_path
        )
    ).wait()
    if return_code != 0:
        raise ValueError(
            f"poetry environment creation in {new_environment_path} failed with return "
            f"code {return_code}"
        )

    # with --no-root, poetry doesn't install wheel or setuptools
    return_code = await (
        await asyncio.create_subprocess_exec(
            ".venv/bin/python",
            "-m",
            "pip",
            "install",
            "setuptools",
            "wheel",
            cwd=new_environment_path,
        )
    ).wait()
    if return_code != 0:
        raise ValueError(
            f"poetry environment creation in {new_environment_path} failed with return "
            f"code {return_code}"
        )



"""
Copied from https://github.com/talkiq/gcloud-aio/tree/auth-4.0.1/auth/gcloud/aio/auth

Applied https://github.com/talkiq/gcloud-aio/pull/518/
"""

__version__ = "4.0.1"

from meadowrun._vendor.gcloud.aio.auth.iam import IamClient
from meadowrun._vendor.gcloud.aio.auth.session import AioSession
from meadowrun._vendor.gcloud.aio.auth.token import Token
from meadowrun._vendor.gcloud.aio.auth.utils import decode
from meadowrun._vendor.gcloud.aio.auth.utils import encode
from meadowrun._vendor.gcloud.aio.auth.build_constants import BUILD_GCLOUD_REST


__all__ = [
    "__version__",
    "IamClient",
    "Token",
    "decode",
    "encode",
    "AioSession",
    "BUILD_GCLOUD_REST",
]



"""
This code cannot reference anything outside of mgmt_functions (as that's what gets
uploaded to the Azure function). We use relative imports which will work both in the
"regular" environment as well as in the Azure function
"""

import asyncio
import dataclasses
import datetime
import json
import logging
from typing import Optional, Sequence, List, Tuple

from ..azure_core.azure_exceptions import ResourceModifiedError
from ..azure_core.azure_rest_api import (
    azure_rest_api,
    azure_rest_api_paged,
    azure_rest_api_poll,
    parse_azure_timestamp,
    wait_for_poll,
)
from ..azure_core.azure_storage_api import (
    StorageAccount,
    azure_table_api,
    azure_table_api_paged,
    table_key_url,
)
from ..azure_constants import (
    LAST_UPDATE_TIME,
    PUBLIC_ADDRESS,
    RUNNING_JOBS,
    SINGLE_PARTITION_KEY,
    VM_ALLOC_TABLE_NAME,
)
from ..mgmt_functions_shared import get_resource_group_path, get_storage_account

# Terminate instances if they haven't run any jobs in the last 5 minutes
_TERMINATE_INSTANCES_IF_IDLE_FOR = datetime.timedelta(minutes=5)
# If we see instances running that aren't registered, we assume there is something wrong
# and they need to be terminated. However, it's possible that we happen to query between
# when an instance is launched and when it's registered. So for the first 5 minutes
# after an instance is launched, we don't terminate it even if it's not registered.
_LAUNCH_REGISTER_DELAY = datetime.timedelta(minutes=5)


@dataclasses.dataclass
class RegisteredVM:
    public_address: str
    last_update_time: datetime.datetime
    num_running_jobs: int
    name: str
    etag: str


@dataclasses.dataclass
class ExistingVM:
    name: str
    time_created: datetime.datetime


async def _get_registered_vms(
    storage_account: StorageAccount,
) -> Sequence[RegisteredVM]:
    """Gets instances registered by AzureInstanceRegistrars"""
    return [
        RegisteredVM(
            item[PUBLIC_ADDRESS],
            datetime.datetime.fromisoformat(item[LAST_UPDATE_TIME]),
            len(json.loads(item[RUNNING_JOBS])),
            item["RowKey"],
            item["odata.etag"],
        )
        async for page in azure_table_api_paged(
            "GET",
            storage_account,
            VM_ALLOC_TABLE_NAME,
            query_parameters={
                "$select": ",".join(
                    ["RowKey", RUNNING_JOBS, LAST_UPDATE_TIME, PUBLIC_ADDRESS]
                )
            },
        )
        for item in page["value"]
    ]


async def _deregister_vm(
    storage_account: StorageAccount, name: str, etag: Optional[str]
) -> bool:
    """
    Deregisters a VM. If etag is None, deregisters unconditionally. If etag is provided,
    only deregisters if the etag matches. Returns False if the etag does not match (i.e.
    optimistic concurrency check failed). Returns True if successful.
    """
    if etag is not None:
        if_match = etag
    else:
        if_match = "*"

    try:
        await azure_table_api(
            "DELETE",
            storage_account,
            table_key_url(VM_ALLOC_TABLE_NAME, SINGLE_PARTITION_KEY, name),
            additional_headers={"If-Match": if_match},
        )
        return True
    except ResourceModifiedError:
        # this is how the API indicates that the etag does not match
        return False


async def _get_running_vms(
    vms_to_check: Sequence[ExistingVM], resource_group_path: str
) -> Tuple[Sequence[ExistingVM], Sequence[ExistingVM]]:
    """Returns (running vms, not running vms)"""

    virtual_machines_path = (
        f"{resource_group_path}/providers/Microsoft.Compute/virtualMachines"
    )

    result: Tuple[List[ExistingVM], List[ExistingVM]] = ([], [])

    for vm in vms_to_check:
        vm_view = await azure_rest_api(
            "GET",
            f"{virtual_machines_path}/{vm.name}",
            "2021-11-01",
            query_parameters={"$expand": "instanceView"},
        )

        power_states = [
            status["code"][len("PowerState/") :]
            for status in vm_view["properties"]["instanceView"]["statuses"]
            if status["code"].startswith("PowerState/")
        ]
        if len(power_states) > 0 and power_states[-1] == "running":
            result[0].append(vm)
        else:
            result[1].append(vm)

    return result


async def _get_all_vms(resource_group_path: str) -> Sequence[ExistingVM]:
    """Gets all VMs via the Azure compute API"""
    return [
        ExistingVM(
            item["name"], parse_azure_timestamp(item["properties"]["timeCreated"])
        )
        async for page in azure_rest_api_paged(
            "GET",
            f"{resource_group_path}/providers/Microsoft.Compute/virtualMachines",
            "2021-11-01",
        )
        for item in page["value"]
    ]


async def _terminate_vm(resource_group_path: str, name: str) -> None:
    await (
        wait_for_poll(
            await azure_rest_api_poll(
                "DELETE",
                f"{resource_group_path}/providers/Microsoft.Compute/virtualMachines/"
                f"{name}",
                "2021-11-01",
                "LocationStatusCode",
            )
        )
    )


async def _deregister_and_terminate_vms(
    storage_account: StorageAccount,
    resource_group_path: str,
    terminate_vms_if_idle_for: datetime.timedelta,
    launch_register_delay: datetime.timedelta = _LAUNCH_REGISTER_DELAY,
) -> List[str]:
    """
    Returns a list of log statements. Because of a bug in Azure Functions, logging calls
    only work from the main thread:
    https://github.com/Azure/azure-functions-python-worker/issues/898 So we return our
    log statements and log them from the main thread, which is unfortunate, but better
    than having no logs.
    """
    logs = []

    all_vms = await _get_all_vms(resource_group_path)
    running_vms, non_running_vms = await _get_running_vms(all_vms, resource_group_path)
    registered_vms = await _get_registered_vms(storage_account)

    now = datetime.datetime.utcnow()

    termination_tasks = []

    existing_vms_names = {vm.name for vm in running_vms}
    for vm in registered_vms:
        if vm.name not in existing_vms_names:
            # deregister VMs that have been registered but don't exist
            logs.append(
                f"{vm.name} ({vm.public_address}) is registered but does not exist, "
                "deregistering"
            )
            await _deregister_vm(storage_account, vm.name, None)
        elif (
            vm.num_running_jobs == 0
            and (now - vm.last_update_time) > terminate_vms_if_idle_for
        ):
            # deregister and terminate machines that have been idle for a while
            success = await _deregister_vm(storage_account, vm.name, vm.etag)
            if success:
                logs.append(
                    f"{vm.name} ({vm.public_address}) is not running any jobs and has "
                    f"not run anything since {vm.last_update_time} so we will "
                    "deregister and terminate it"
                )

                termination_tasks.append(
                    asyncio.create_task(_terminate_vm(resource_group_path, vm.name))
                )
        else:
            logs.append(
                f"Letting {vm.name} ({vm.public_address}) continue to run--this VM is "
                "active"
            )

    registered_vms_names = {vm.name for vm in registered_vms}
    for vm in running_vms:
        if vm.name not in registered_vms_names:
            if (now - vm.time_created) > launch_register_delay:
                # terminate instances that are running but not registered
                logs.append(
                    f"{vm.name} is running but not registered, will terminate. Was "
                    f"launched at {vm.time_created}"
                )
                termination_tasks.append(
                    asyncio.create_task(_terminate_vm(resource_group_path, vm.name))
                )
            else:
                print(
                    f"{vm.name} is running but is not registered, not terminating "
                    f"because it was launched recently at {vm.time_created}"
                )

    if termination_tasks:
        await asyncio.wait(termination_tasks)

    return logs


async def adjust() -> List[str]:
    """
    This code is designed to run on an Azure function as set up by
    azure_mgmt_functions_setup
    """
    return await _deregister_and_terminate_vms(
        get_storage_account(),
        get_resource_group_path(),
        _TERMINATE_INSTANCES_IF_IDLE_FOR,
        _LAUNCH_REGISTER_DELAY,
    )


async def terminate_all_vms(resource_group_path: str) -> None:
    terminate_tasks = []
    for vm in await _get_all_vms(resource_group_path):
        print(f"Terminating {vm.name}")
        terminate_tasks.append(
            asyncio.create_task(_terminate_vm(resource_group_path, vm.name))
        )
    if terminate_tasks:
        await asyncio.wait(terminate_tasks)


def main(myTimer) -> None:  # type: ignore
    """
    The entry point for the Azure function (see function.json). Configured to run every
    minute by default.

    myTime should be annotated as azure.functions.TimerRequest, but we'd rather not add
    that dependency (it exists by default in the function runtime where this function
    will actually run). Also, the variable cannot be snake_case, and can only be changed
    if the corresponding name in function.json is changed.
    """
    logs = asyncio.run(adjust())
    for log in logs:
        logging.info(log)



from __future__ import annotations

import os
import pickle
from typing import TYPE_CHECKING, Any, Dict, Optional, Tuple, Union

if TYPE_CHECKING:
    from typing_extensions import Literal, Final

_UNINITIALIZED: Final = "__UNINITIALIZED__"


_MEADOWRUN_CONTEXT_VARIABLES = "MEADOWRUN_CONTEXT_VARIABLES"

_variables: Optional[Dict[str, Any]] = None


def variables() -> Dict[str, Any]:
    global _variables

    if _variables is None:
        if _MEADOWRUN_CONTEXT_VARIABLES in os.environ:
            with open(os.environ[_MEADOWRUN_CONTEXT_VARIABLES], "rb") as f:
                # TODO probably should provide nicer error messages on unpickling
                _variables = pickle.load(f)
        else:
            _variables = {}

    return _variables


_MEADOWRUN_RESULT_FILE = "MEADOWRUN_RESULT_FILE"
_MEADOWRUN_RESULT_PICKLE_PROTOCOL = "MEADOWRUN_RESULT_PICKLE_PROTOCOL"

# We use a non-None placeholder value because we want to use None to mean that these
# variables were not set
_result_request: Optional[
    Union[Literal["__UNINITIALIZED__"], Tuple[str, int]]
] = _UNINITIALIZED


def result_request() -> Optional[Tuple[str, int]]:
    """
    If this returns non-None, the result is (result file, pickle protocol). If the
    result is non-None, that means that the current process was launched by meadowrun in
    a way that meadowrun wasn't able to wrap the process to guarantee that results get
    written. As a result, we should do our best to write results to the specified file
    using the specified pickle protocol version.
    """
    global _result_request

    if _result_request == _UNINITIALIZED:
        if _MEADOWRUN_RESULT_FILE in os.environ:
            file = os.environ[_MEADOWRUN_RESULT_FILE]
            # We should never have _MEADOWGRID_RESULT_FILE without
            # _MEADOWGRID_RESULT_PICKLE_PROTOCOL which should always be an integer.
            # Instead of failing if this is the case, though, we just assume a
            # (relatively) low pickle protocol
            protocol = 3  # the fallback protocol version
            if _MEADOWRUN_RESULT_PICKLE_PROTOCOL in os.environ:
                protocol_str = os.environ[_MEADOWRUN_RESULT_PICKLE_PROTOCOL]
                try:
                    protocol = int(protocol_str)
                except ValueError:
                    # ignore error and use fallback
                    pass
            _result_request = file, protocol
        else:
            _result_request = None

    return _result_request



from __future__ import annotations

import asyncio
import uuid
from typing import TYPE_CHECKING, Optional, Tuple, Sequence, Awaitable, Any, Dict

if TYPE_CHECKING:
    from typing_extensions import Literal

from meadowrun.azure_integration.azure_meadowrun_core import (
    _ensure_managed_identity,
    ensure_meadowrun_resource_group,
    get_default_location,
)
from meadowrun.azure_integration.azure_vm_pricing import get_vm_types
from meadowrun.azure_integration.mgmt_functions.azure_core.azure_exceptions import (
    ResourceNotFoundError,
)
from meadowrun.azure_integration.mgmt_functions.azure_core.azure_rest_api import (
    azure_rest_api,
    azure_rest_api_poll,
    wait_for_poll,
)
from meadowrun.instance_selection import (
    CloudInstance,
    OnDemandOrSpotType,
    ResourcesInternal,
    choose_instance_types_for_job,
)
from meadowrun.version import __version__


_MEADOWRUN_VIRTUAL_NETWORK_NAME = "Meadowrun-vnet"
_MEADOWRUN_SUBNET_NAME = "Meadowrun-subnet"
_MEADOWRUN_USERNAME = "meadowrunuser"


# To get this ID, build an image using build_scripts/build_azure_image.py. Once you have
# an image, you'll need to get the community image id (NOT the image's regular id!) and
# copy/paste that here, using the following command:
# https://docs.microsoft.com/en-us/cli/azure/sig/image-definition?view=azure-cli-latest#az-sig-image-definition-list-community
_MEADOWRUN_COMMUNITY_IMAGE_ID = (
    "/CommunityGalleries/meadowprodeastus-e8b60fd5-8978-467b-a1b0-5b83cbf5393d/Images/"
    "meadowrun"
)


async def _ensure_virtual_network_and_subnet(location: str) -> str:
    """
    Returns the meadowrun subnet's id, creates it if it doesn't exist, and creates the
    meadowrun virtual network if that doesn't exist either.
    """
    virtual_network_path = (
        f"{await ensure_meadowrun_resource_group(location)}/providers/"
        f"Microsoft.Network/virtualnetworks/{_MEADOWRUN_VIRTUAL_NETWORK_NAME}"
    )
    subnet_path = f"{virtual_network_path}/subnets/{_MEADOWRUN_SUBNET_NAME}"

    try:
        subnet = await azure_rest_api("GET", subnet_path, "2021-05-01")
    except ResourceNotFoundError:
        print(
            f"The meadowrun virtual network ({_MEADOWRUN_VIRTUAL_NETWORK_NAME}) and/or "
            f"the subnet in it ({_MEADOWRUN_SUBNET_NAME}) does not exist, creating it "
            f"now in {location}"
        )

        # Provision the virtual network and wait for completion, will just get updated
        # if it already exists.
        # https://docs.microsoft.com/en-us/rest/api/virtualnetwork/virtual-networks/create-or-update
        await wait_for_poll(
            await azure_rest_api_poll(
                "PUT",
                virtual_network_path,
                "2021-05-01",
                "AsyncOperationJsonStatus",
                json_content={
                    "location": location,
                    "properties": {
                        "addressSpace": {"addressPrefixes": ["10.0.0.0/16"]},
                    },
                },
            )
        )

        # Then create the subnet
        # https://docs.microsoft.com/en-us/rest/api/virtualnetwork/subnets/create-or-update
        subnet, _ = await wait_for_poll(
            await azure_rest_api_poll(
                "PUT",
                subnet_path,
                "2021-05-01",
                "AsyncOperationJsonStatus",
                json_content={"properties": {"addressPrefix": "10.0.0.0/24"}},
            )
        )

    return subnet["id"]


async def _provision_nic_with_public_ip(
    location: str,
    vm_name: str,
) -> Tuple[str, str, Optional[Awaitable[Any]], Optional[Awaitable[Any]]]:
    """
    Creates a public IP address and corresponding NIC. Returns (NIC id, public address
    path, NIC creation continuation, ip address creation continuation). The NIC id and
    public address path can be used to start provisioning a VM, but the continuations
    must be awaited before using the VM
    """

    subnet_id = await _ensure_virtual_network_and_subnet(location)

    # provision an IP address
    # TODO there should probably be a way to re-use IP addresses rather than
    # create/destroy them for every VM
    resource_group_path = await ensure_meadowrun_resource_group(location)
    public_ip_path = (
        f"{resource_group_path}/providers/Microsoft.Network/publicIPAddresses"
        f"/{vm_name}-ip"
    )

    # https://docs.microsoft.com/en-us/rest/api/virtualnetwork/public-ip-addresses/create-or-update
    ip_address_result, ip_address_continuation = await azure_rest_api_poll(
        "PUT",
        public_ip_path,
        "2021-05-01",
        "AsyncOperationJsonStatus",
        json_content={
            "location": location,
            # critical that we use Basic here instead of Standard, otherwise this will
            # not be available via instance metadata (IMDS):
            # https://github.com/MicrosoftDocs/azure-docs/issues/44314
            "sku": {"name": "Basic"},
            "properties": {
                "publicIPAllocationMethod": "Dynamic",
                "publicIPAddressVersion": "IPV4",
            },
        },
    )

    # provision a NIC
    nic_result, nic_continuation = await azure_rest_api_poll(
        "PUT",
        f"{resource_group_path}/providers/Microsoft.Network/networkInterfaces"
        f"/{vm_name}-nic",
        "2021-05-01",
        "AsyncOperationJsonStatus",
        json_content={
            "location": location,
            "properties": {
                "ipConfigurations": [
                    {
                        "name": f"{vm_name}-ip-config",
                        "properties": {
                            "subnet": {"id": subnet_id},
                            "publicIPAddress": {
                                "id": ip_address_result["id"],
                                "properties": {"deleteOption": "Delete"},
                            },
                        },
                    }
                ]
            },
        },
    )

    return nic_result["id"], public_ip_path, nic_continuation, ip_address_continuation


async def _provision_vm(
    location: str,
    vm_size: str,
    on_demand_or_spot: OnDemandOrSpotType,
    ssh_public_key_data: str,
    image_id: str,
    image_type: Literal["Community", "Regular"],
) -> Tuple[str, str]:
    """
    Based on
    https://docs.microsoft.com/en-us/azure/developer/python/sdk/examples/azure-sdk-example-virtual-machines?tabs=cmd

    Returns the (public IP address of the VM, name of the VM).

    ssh_public_key_data should be the actual line that should get added to
    ~/.ssh/authorized_keys, e.g. "ssh-rsa: ...". Azure does seem to have a feature for
    storing SSH public keys for reuse, but there doesn't seem to be a way to reference
    them from the VM API.
    """

    resource_group_path = await ensure_meadowrun_resource_group(location)
    meadowrun_identity_id, _ = await _ensure_managed_identity(location)

    vm_name = str(uuid.uuid4())

    # every VM needs its own NIC and IP address
    (
        nic_id,
        ip_address_path,
        nic_continuation,
        ip_address_continuation,
    ) = await _provision_nic_with_public_ip(location, vm_name)

    # https://docs.microsoft.com/en-us/rest/api/compute/virtual-machines/create-or-update
    # The documentation here is poor, the best way to understand how to use these
    # options is to create VMs using the portal and then run `az vm list` on the
    # command line to see how the portal sets these options.
    vm_create_properties: Dict[str, Any] = {
        "location": location,
        "identity": {
            "type": "UserAssigned",
            "userAssignedIdentities": {meadowrun_identity_id: {}},
        },
        "properties": {
            "hardwareProfile": {
                "vmSize": vm_size,
            },
            "storageProfile": {
                "imageReference": {
                    # will be populated further down
                },
                "osDisk": {
                    "createOption": "FromImage",
                    "deleteOption": "Delete",
                },
            },
            "osProfile": {
                "computerName": f"{vm_name}-vm",
                "adminUsername": _MEADOWRUN_USERNAME,
                "linuxConfiguration": {
                    "disablePasswordAuthentication": True,
                    "ssh": {
                        "publicKeys": [
                            {
                                "path": (
                                    f"/home/{_MEADOWRUN_USERNAME}/.ssh/authorized_keys"
                                ),
                                "keyData": ssh_public_key_data,
                            }
                        ]
                    },
                },
            },
            "networkProfile": {
                "networkInterfaces": [
                    {"id": nic_id, "properties": {"deleteOption": "Delete"}}
                ]
            },
        },
    }
    if on_demand_or_spot == "spot":
        vm_create_properties["properties"]["priority"] = "Spot"

    if image_type == "Community":
        vm_create_properties["properties"]["storageProfile"]["imageReference"][
            "communityGalleryImageId"
        ] = image_id
    elif image_type == "Regular":
        vm_create_properties["properties"]["storageProfile"]["imageReference"][
            "id"
        ] = image_id
    else:
        raise ValueError(f"Unexpected value for image_type {image_type}")

    vm_result, vm_continuation = await azure_rest_api_poll(
        "PUT",
        f"{resource_group_path}/providers/Microsoft.Compute/virtualMachines/"
        f"{vm_name}-vm",
        "2021-11-01",
        "AsyncOperationJsonStatus",
        json_content=vm_create_properties,
    )

    continuations = [
        continuation
        for continuation in (nic_continuation, ip_address_continuation, vm_continuation)
        if continuation is not None
    ]
    if continuations:
        await asyncio.gather(*continuations)

    # we have to re-get the ip address object to get the current ip address, as it
    # doesn't get assigned until the ip address is attached to a VM and the VM is
    # launched
    ip_address_result2 = await azure_rest_api("GET", ip_address_path, "2021-05-01")
    ip_address = ip_address_result2["properties"]["ipAddress"]

    print(f"Provisioned virtual machine {vm_result['name']} ({ip_address})")

    # TODO the "deleteOption: Delete" takes care of deleting disks, network
    # interfaces, and public IP addresses when the VM is deleted, but we also need
    # to worry about the case where we throw an exception and/or crash after those
    # resources are created but before the VM is created

    return ip_address, vm_result["name"]


async def launch_vms(
    resources_required_per_job: ResourcesInternal,
    num_jobs: int,
    ssh_public_key_data: str,
    location: Optional[str] = None,
) -> Sequence[CloudInstance]:
    """See docstring for launch_ec2_instances"""

    if location is None:
        location = get_default_location()

    chosen_instance_types = choose_instance_types_for_job(
        resources_required_per_job,
        num_jobs,
        await get_vm_types(location),
    )

    vm_detail_tasks = []
    chosen_instance_types_repeated = []

    image_id = f"{_MEADOWRUN_COMMUNITY_IMAGE_ID}/versions/{__version__}"
    # for development, you can uncomment the following, and replace "Community" with
    # "Regular" below
    # image_id = (
    #     "/subscriptions/d740513f-4172-4792-bd29-5194e79d5881/resourceGroups/"
    #     "meadowrun-dev/providers/Microsoft.Compute/galleries/meadowrun.dev.gallery/"
    #     f"images/meadowrun/versions/{_MEADOWRUN_IMAGE_VERSION}"
    # )

    at_least_one_chosen_instance_type = False
    for instance_type in chosen_instance_types:
        at_least_one_chosen_instance_type = True
        for _ in range(instance_type.num_instances):
            vm_detail_tasks.append(
                _provision_vm(
                    location,
                    instance_type.instance_type.name,
                    instance_type.instance_type.on_demand_or_spot,
                    ssh_public_key_data,
                    image_id,
                    "Community",
                )
            )
            chosen_instance_types_repeated.append(instance_type)

    if not at_least_one_chosen_instance_type:
        raise ValueError(
            "There were no instance types that could be selected for "
            f"{resources_required_per_job}"
        )

    vm_details = await asyncio.gather(*vm_detail_tasks)

    return [
        CloudInstance(public_dns_name, vm_name, instance_type)
        for (public_dns_name, vm_name), instance_type in zip(
            vm_details, chosen_instance_types_repeated
        )
    ]



"""
There's no such thing as an "S3 grid job". The code in this module just helps us
implement grid jobs that use an S3-compatible object store to transfer data, e.g. for
EC2 and Kubernetes
"""


from __future__ import annotations

import asyncio
import hashlib
import io
import json
import os
import pickle
import shutil
import tempfile
import time
from typing import (
    Any,
    AsyncIterable,
    Iterable,
    List,
    Optional,
    Set,
    TYPE_CHECKING,
    Tuple,
    Type,
)

import aiobotocore.session
import boto3.exceptions
import botocore.exceptions
from botocore.exceptions import ClientError

from meadowrun._vendor.fastcdc.fastcdc_py import fastcdc_py
from meadowrun.abstract_storage_bucket import AbstractStorageBucket
from meadowrun.aws_integration.aws_core import (
    MeadowrunAWSAccessError,
    MeadowrunNotInstalledError,
    get_bucket_name,
)
from meadowrun.meadowrun_pb2 import ProcessState
from meadowrun.run_job_core import (
    JobCompletion,
    MeadowrunException,
    TaskProcessState,
    WorkerProcessState,
)
from meadowrun.storage_keys import (
    STORAGE_KEY_PROCESS_STATE_SUFFIX,
    STORAGE_KEY_TASK_RESULT_SUFFIX,
    parse_storage_key_process_state,
    parse_storage_key_task_result,
    storage_key_process_state,
    storage_key_task_args,
    storage_key_task_result,
    storage_prefix_outputs,
)

if TYPE_CHECKING:
    import types_aiobotocore_s3
    from typing_extensions import Literal
    from types import TracebackType


class GenericStorageBucket(AbstractStorageBucket):
    """
    This should be used for accessing AWS S3 as well as any other
    username/password-based S3-compatible object storage systems (e.g. Minio)
    """

    def __init__(
        self, s3_client: types_aiobotocore_s3.S3Client, bucket: str, cache_key: str
    ):
        self._s3_client = s3_client
        self._bucket = bucket
        self._cache_key = cache_key

    async def __aenter__(self) -> GenericStorageBucket:
        self._s3_client = await self._s3_client.__aenter__()
        return self

    async def __aexit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> None:
        await self._s3_client.__aexit__(exc_type, exc_val, exc_tb)

    def get_cache_key(self) -> str:
        return self._cache_key

    async def get_bytes(self, key: str) -> bytes:
        response = await self._s3_client.get_object(Bucket=self._bucket, Key=key)
        async with response["Body"] as stream:
            return await stream.read()

    async def try_get_bytes(self, key: str) -> Optional[bytes]:
        try:
            return await self.get_bytes(key)
        except botocore.exceptions.ClientError as error:
            if error.response["Error"]["Code"] not in ("404", "NoSuchKey"):
                raise
            return None

    async def get_byte_range(self, key: str, byte_range: Tuple[int, int]) -> bytes:
        response = await self._s3_client.get_object(
            Bucket=self._bucket, Key=key, Range=f"bytes={byte_range[0]}-{byte_range[1]}"
        )

        async with response["Body"] as stream:
            return await stream.read()

    async def write_bytes(self, data: bytes, key: str) -> None:
        await self._s3_client.put_object(Bucket=self._bucket, Key=key, Body=data)

    async def exists(self, key: str) -> bool:
        try:
            await self._s3_client.head_object(Bucket=self._bucket, Key=key)
            return True
        except ClientError as error:
            # don't raise an error saying the file doesn't exist
            if error.response["Error"]["Code"] not in ("404", "NoSuchKey"):
                raise error
        return False

    async def get_file(self, key: str, local_filename: str) -> None:
        response = await self._s3_client.get_object(Bucket=self._bucket, Key=key)
        async with response["Body"] as stream:
            with open(local_filename, "wb") as f:
                f.write(await stream.read())

    async def try_get_file(self, key: str, local_filename: str) -> bool:
        try:
            await self.get_file(key, local_filename)
            return True
        except botocore.exceptions.ClientError as error:
            # don't raise an error saying the file doesn't exist
            if error.response["Error"]["Code"] not in ("404", "NoSuchKey"):
                raise error

            return False

    async def write_file(self, local_filename: str, key: str) -> None:
        with open(local_filename, "rb") as f:
            await self._s3_client.put_object(Bucket=self._bucket, Key=key, Body=f)

    async def list_objects(self, key_prefix: str) -> List[str]:
        """Returns keys with the matching prefix in the bucket"""
        paginator = self._s3_client.get_paginator("list_objects_v2")
        results = []
        async for result in paginator.paginate(Bucket=self._bucket, Prefix=key_prefix):
            for c in result.get("Contents", []):
                results.append(c["Key"])
        return results

    async def delete_object(self, key: str) -> None:
        await self._s3_client.delete_object(Bucket=self._bucket, Key=key)


class S3Bucket(GenericStorageBucket):
    """
    This should be used for accessing AWS S3 as well as any other
    username/password-based S3-compatible object storage systems (e.g. Minio)
    """

    def __init__(
        self, s3_client: types_aiobotocore_s3.S3Client, bucket: str, cache_key: str
    ):
        super().__init__(s3_client, bucket, cache_key)

    async def __aenter__(self) -> S3Bucket:
        # annoying, only needed for type checking
        await super().__aenter__()
        return self

    async def write_bytes_if_not_exists(self, data: bytes, key: str) -> None:
        # this override just provides some nicer error messages. This function is
        # usually called before other functions
        try:
            await super().write_bytes_if_not_exists(data, key)
        except boto3.exceptions.S3UploadFailedError as e:
            if len(e.args) >= 1 and "NoSuchBucket" in e.args[0]:
                raise MeadowrunNotInstalledError("S3 bucket")
            raise
        except ClientError as error:
            if error.response["Error"]["Code"] == "403":
                # if we don't have permissions to the bucket, throw a helpful error
                raise MeadowrunAWSAccessError("S3 bucket") from error
            raise


def get_generic_username_password_bucket(
    storage_endpoint_url: str,
    storage_access_key_id: Optional[str],
    storage_secret_access_key: Optional[str],
    storage_bucket: str,
) -> GenericStorageBucket:
    """
    This can be used with any S3-compatible object storage system that requires a
    username/password
    """
    kwargs = {}
    if storage_access_key_id is not None:
        kwargs["aws_access_key_id"] = storage_access_key_id
    if storage_secret_access_key is not None:
        kwargs["aws_secret_access_key"] = storage_secret_access_key

    session = aiobotocore.session.get_session()
    s3_client = session.create_client(  # type: ignore
        "s3", endpoint_url=storage_endpoint_url, **kwargs
    )
    return GenericStorageBucket(
        s3_client, storage_bucket, f"{storage_endpoint_url}/{storage_bucket}"
    )


def get_aws_s3_bucket(region_name: str) -> S3Bucket:
    """
    This is for S3 clients where we are connecting to the real S3 using the default
    Meadowrun infrastructure for EC2/AWS. In that case, the name of the bucket we
    use can be constructed from the region name. This function should NOT be used if
    we are using an S3 client pointing to a non-S3 object storage.
    """
    bucket_name = get_bucket_name(region_name)
    return S3Bucket(
        aiobotocore.session.get_session().create_client("s3", region_name=region_name),
        bucket_name,
        f"s3/{bucket_name}",
    )


async def get_aws_s3_bucket_async(region_name: str) -> S3Bucket:
    """
    A bit silly, but sometimes we want an Awaitable[S3Bucket] rather than an S3Bucket
    because other StorageBucket constructors are async
    """
    return get_aws_s3_bucket(region_name)


async def upload_task_args(
    storage_bucket: AbstractStorageBucket,
    job_id: str,
    args: Iterable[Any],
) -> List[Tuple[int, int]]:
    range_from = 0
    ranges = []
    with io.BytesIO() as buffer:
        for i, arg in enumerate(args):
            pickle.dump(((arg,), {}), buffer)
            range_to = buffer.tell() - 1
            ranges.append((range_from, range_to))
            range_from = range_to + 1

        await storage_bucket.write_bytes(
            buffer.getvalue(), storage_key_task_args(job_id)
        )

    return ranges


async def download_task_arg(
    storage_bucket: AbstractStorageBucket,
    job_id: str,
    byte_range: Tuple[int, int],
) -> Any:
    return await storage_bucket.get_byte_range(
        storage_key_task_args(job_id), byte_range
    )


async def complete_task(
    storage_bucket: AbstractStorageBucket,
    job_id: str,
    task_id: int,
    attempt: int,
    process_state: ProcessState,
) -> None:
    """Uploads the result of the task to S3."""
    await storage_bucket.write_bytes(
        process_state.SerializeToString(),
        storage_key_task_result(job_id, task_id, attempt),
    )


async def receive_results(
    storage_bucket: AbstractStorageBucket,
    job_id: str,
    stop_receiving: asyncio.Event,
    all_workers_exited: asyncio.Event,
    initial_wait_seconds: int = 1,
    receive_message_wait_seconds: int = 20,
    read_worker_process_states: bool = True,
) -> AsyncIterable[Tuple[List[TaskProcessState], List[WorkerProcessState]]]:
    """
    Listens to a result queue until we have results for num_tasks.

    As results become available, yields (task results, worker results). Task results
    will be a list of TaskProcessState. Worker results will be a list of
    WorkerProcessState
    """

    # Behavior is that if stop_receiving is set, we want to return immediately. If
    # all_workers_exited is set, then keep trying for about 3 seconds (just in case some
    # results are still coming in), and then return

    delete_tasks = []

    try:
        results_prefix = storage_prefix_outputs(job_id)
        download_keys_received: Set[str] = set()
        wait = initial_wait_seconds
        workers_exited_wait_count = 0
        while not stop_receiving.is_set() and (
            workers_exited_wait_count < 3 or wait == 0
        ):
            if all_workers_exited.is_set():
                workers_exited_wait_count += 1

            if wait:
                events_to_wait_for = [asyncio.create_task(stop_receiving.wait())]
                if workers_exited_wait_count == 0:
                    events_to_wait_for.append(
                        asyncio.create_task(all_workers_exited.wait())
                    )
                else:
                    # poll more frequently if workers are done, but still wait 1 second
                    # (unless stop_receiving is set)
                    wait = 1

                done, pending = await asyncio.wait(
                    events_to_wait_for,
                    timeout=wait,
                    return_when=asyncio.FIRST_COMPLETED,
                )
                for p in pending:
                    p.cancel()
                if stop_receiving.is_set():
                    break

            keys = await storage_bucket.list_objects(results_prefix)

            download_tasks = []
            for key in keys:
                if key not in download_keys_received:
                    download_tasks.append(
                        asyncio.create_task(storage_bucket.get_bytes_and_key(key))
                    )

            download_keys_received.update(keys)

            if len(download_tasks) == 0:
                if wait == 0:
                    wait = 1
                else:
                    wait = min(wait + 1, receive_message_wait_seconds)
            else:
                wait = 0
                task_results = []
                worker_results = []
                for task_result_future in asyncio.as_completed(download_tasks):
                    process_state_bytes, key = await task_result_future
                    process_state = ProcessState()
                    process_state.ParseFromString(process_state_bytes)
                    if key.endswith(STORAGE_KEY_TASK_RESULT_SUFFIX):
                        task_id, attempt = parse_storage_key_task_result(
                            key, results_prefix
                        )
                        task_results.append(
                            TaskProcessState(task_id, attempt, process_state)
                        )
                    elif key.endswith(STORAGE_KEY_PROCESS_STATE_SUFFIX):
                        if not read_worker_process_states:
                            continue

                        worker_index = parse_storage_key_process_state(
                            key, results_prefix
                        )
                        worker_results.append(
                            WorkerProcessState(worker_index, process_state)
                        )
                    else:
                        print(f"Warning, unrecognized key {key}, will ignore")
                        continue  # don't delete if we can't parse the key

                    delete_tasks.append(
                        asyncio.create_task(storage_bucket.delete_object(key))
                    )
                yield task_results, worker_results
    finally:
        await asyncio.gather(*delete_tasks, return_exceptions=True)


async def get_job_completion_from_process_state(
    storage_bucket: AbstractStorageBucket,
    job_id: str,
    worker_index: str,
    job_spec_type: Literal["py_command", "py_function", "py_agent"],
    timeout_seconds: int,
    public_address: str,
) -> JobCompletion[Any]:
    """
    Polls for the specified .process_state file, and then creates a JobCompletion object
    based on the process_state

    file_suffix is used by indexed completion jobs to distinguish between the
    completions of different workers, this should be set to the job completion index.
    """

    t0 = time.time()
    process_state_bytes = None
    process_state_key = storage_key_process_state(job_id, worker_index)
    wait = 1
    while time.time() - t0 < timeout_seconds:
        process_state_bytes = await storage_bucket.try_get_bytes(process_state_key)
        if process_state_bytes is not None:
            break
        await asyncio.sleep(wait)
        wait = max(wait + 1, 20)

    if process_state_bytes is None:
        raise TimeoutError(
            f"Waited {timeout_seconds} seconds but {process_state_key} does not exist"
        )

    process_state = ProcessState.FromString(process_state_bytes)
    if process_state.state == ProcessState.ProcessStateEnum.SUCCEEDED:
        # we must have a result from functions, in other cases we can optionally have a
        # result
        if job_spec_type == "py_function" or process_state.pickled_result:
            result = pickle.loads(process_state.pickled_result)
        else:
            result = None
        return JobCompletion(
            result,
            process_state.state,
            process_state.log_file_name,
            process_state.return_code,
            public_address,
        )
    else:
        raise MeadowrunException(process_state)


# This is ugly, but avoids repeated checking if the same bucket+key exists. Only used in
# ensure_uploaded_incremental
_existing_keys: Set[Tuple[str, str]] = set()
_CHUNKS_KEY = "chunks"


async def ensure_uploaded_incremental(
    storage_bucket: AbstractStorageBucket,
    local_file_path: str,
    key_prefix: str = "",
    avg_chunk_size: int = 1000_000,
) -> str:
    """
    Uploads the specified file from the local machine to S3, in chunks. Each chunk is
    content-hashed and uploaded to "{key_prefix}{hash}.part", where {hash} is a hash of
    the contents of the file. A JSON file with all the parts is uploaded to the key
    "{key_prefix}{hash}.json", where {hash} is a hash of the contents.

    The file is split using content-based-chunking - this means there's no exact size
    for each chunk, but the size can be controlled by setting avg_chunk_size.

    If any S3 object already exists, it is not re-uploaded.

    key_prefix should usually be "" or end in a "/" like "code/".

    Returns the key of the JSON file in the S3 bucket, e.g. "code/123456789abcdefg.json"
    """

    # TODO: only check for existence of chunks if chunks_json does not exist?

    chunk_upload_tasks = []
    keys = []
    for chunk in fastcdc_py(
        local_file_path, avg_size=avg_chunk_size, fat=True, hf=hashlib.blake2b
    ):
        key = f"{key_prefix}{chunk.hash}.part"
        keys.append(key)
        cache_key = (storage_bucket.get_cache_key(), key)
        if (storage_bucket.get_cache_key(), key) not in _existing_keys:
            chunk_upload_tasks.append(
                asyncio.create_task(
                    storage_bucket.write_bytes_if_not_exists(chunk.data, key)
                )
            )
            _existing_keys.add(cache_key)

    results = await asyncio.gather(*chunk_upload_tasks, return_exceptions=True)
    for result in results:
        if isinstance(result, Exception):
            # on exceptions, we're not sure what the state of the uploaded keys is.
            _existing_keys.clear()
            raise result

    chunks_json = json.dumps({_CHUNKS_KEY: keys}).encode("utf-8")
    chunks_json_key = f"{key_prefix}{hashlib.blake2b(chunks_json).hexdigest()}.json"
    await storage_bucket.write_bytes_if_not_exists(chunks_json, chunks_json_key)

    return chunks_json_key


async def download_chunked_file(
    storage_bucket: AbstractStorageBucket,
    object_name: str,
    file_name: str,
) -> None:
    """Download a file that was uploaded in chunks via ensure_uploaded_incremental"""
    chunk_list_bs = await storage_bucket.get_bytes(object_name)
    chunks = json.loads(chunk_list_bs.decode("utf-8"))[_CHUNKS_KEY]

    with tempfile.TemporaryDirectory() as tmp:
        await _download_files(storage_bucket, chunks, tmp)
        _concat_chunks(tmp, chunks, file_name)


def _concat_chunks(
    source_folder: str, chunks: Iterable[str], destination_file: str
) -> None:
    with open(destination_file, "wb") as dest:
        for source_file in chunks:
            with open(
                os.path.join(source_folder, os.path.basename(source_file)), "rb"
            ) as src:
                shutil.copyfileobj(src, dest)


async def _download_files(
    storage_bucket: AbstractStorageBucket,
    keys: Iterable[str],
    target_folder: str,
) -> None:
    """Download all given keys in the bucket as separate files to the target folder, in
    parallel. Keys can contain duplicates, they'll be downloaded only once.
    """
    download_tasks = []
    for key in set(keys):
        download_tasks.append(
            asyncio.create_task(
                storage_bucket.get_file(
                    key, os.path.join(target_folder, os.path.basename(key))
                )
            )
        )
    await asyncio.gather(*download_tasks)



"""
The sole purpose of this script is to run until this pod has been idle for at least
_IDLE_TIMEOUT_SECS, and then exit at that point. Actual jobs will be started using
kubectl exec on this pod.

Not dissimilar to deallocate_jobs.py
"""
import time
import traceback

from meadowrun.k8s_integration.is_job_running import is_job_running

_LAST_JOB_TIMESTAMP_FILE = "/var/meadowrun/job_timestamp"

_IDLE_TIMEOUT_SECS = 60 * 5


def main() -> None:
    last_activity = time.time()
    while True:
        if time.time() - last_activity > _IDLE_TIMEOUT_SECS:
            print(
                f"Last job was running at {last_activity}, which is more than "
                f"{_IDLE_TIMEOUT_SECS} ago, exiting"
            )
            break

        time.sleep(30)

        try:
            # this checks for jobs that are currently running
            if is_job_running():
                last_activity = time.time()
            else:
                # this checks for jobs that ran and exited quickly that we may not have
                # "noticed" were running
                with open(_LAST_JOB_TIMESTAMP_FILE, "r", encoding="utf-8") as f:
                    last_job_timestamp = float(f.read())
                if last_job_timestamp > last_activity:
                    last_activity = last_job_timestamp
        except BaseException:
            print("Error checking for running jobs: " + traceback.format_exc())


if __name__ == "__main__":
    main()



from __future__ import annotations

import abc
import asyncio
import dataclasses
import itertools
import uuid
from typing import (
    TYPE_CHECKING,
    Any,
    AsyncIterable,
    Dict,
    Generic,
    Iterable,
    List,
    Optional,
    Sequence,
    Tuple,
    Type,
    TypeVar,
)

from meadowrun.instance_selection import (
    CloudInstance,
    ResourcesInternal,
    remaining_resources_sort_key,
)
from meadowrun.shared import assert_is_not_none

if TYPE_CHECKING:
    from types import TracebackType
    from meadowrun.run_job_core import AllocVM


@dataclasses.dataclass
class _InstanceState:
    """Represents an existing instance"""

    public_address: str

    name: str

    available_resources: Optional[ResourcesInternal]

    running_jobs: Optional[Dict[str, Dict[str, Any]]]

    prevent_further_allocation: bool

    def get_available_resources(self) -> ResourcesInternal:
        """
        The current available resources on the instance. This will change as jobs get
        allocated/deallocated to the instance.

        We use this function instead of directly accessing the property as a bit of a
        hack to get some duck-typing so that we only need to populate
        available_resources when it will actually be used. Pulling these properties when
        they're not needed is potentially expensive, as that data comes from DynamoDB or
        an Azure Table. Functions that return _InstanceState should document whether
        this property will be available.
        """
        if self.available_resources is None:
            raise ValueError(
                "Programming error--_InstanceState.available_resources was requested "
                "but is None"
            )
        return self.available_resources

    def get_running_jobs(self) -> Dict[str, Dict[str, Any]]:
        """
        Returns job_id -> metadata, where metadata has keys _LOGICAL_CPU_ALLOCATED,
        _MEMORY_GB_ALLOCATED, _ALLOCATED_TIME

        Also see comment on get_available_resources
        """
        if self.running_jobs is None:
            raise ValueError(
                "Programming error--_InstanceState.get_running_jobs was requested but "
                "is None"
            )
        return self.running_jobs


_TInstanceState = TypeVar("_TInstanceState", bound=_InstanceState)


class InstanceRegistrar(abc.ABC, Generic[_TInstanceState]):
    """
    An implementation of InstanceRegistrar provides a way to register "instances" (e.g.
    EC2 instances or Azure VMs) as they're created, and then allocate/deallocate jobs to
    these instances. Another way of thinking about it is a serverless cluster manager.
    """

    @abc.abstractmethod
    async def __aenter__(self) -> InstanceRegistrar:
        """Initializes the InstanceRegistrar"""
        pass

    @abc.abstractmethod
    async def __aexit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> None:
        pass

    @abc.abstractmethod
    def get_region_name(self) -> str:
        pass

    @abc.abstractmethod
    async def register_instance(
        self,
        public_address: str,
        name: str,
        resources_available: ResourcesInternal,
        running_jobs: List[Tuple[str, ResourcesInternal]],
    ) -> None:
        """Registers a (presumably newly created) instance"""
        pass

    @abc.abstractmethod
    async def get_registered_instances(self) -> List[_TInstanceState]:
        """
        Gets all registered instances. Must have available_resources populated.
        running_jobs is optional depending on whether the corresponding implementation
        of allocate_jobs_to_instance will need it.
        """
        pass

    @abc.abstractmethod
    async def get_registered_instance(self, name: str) -> _TInstanceState:
        """
        Gets the InstanceRegistrar's representation of the specified instance.

        Must populate running_jobs, available_resources is optional depending on whether
        the corresponding implementation of deallocate_job_from_instance will need it
        """
        pass

    @abc.abstractmethod
    async def allocate_jobs_to_instance(
        self,
        instance: _TInstanceState,
        resources_allocated_per_job: ResourcesInternal,
        new_job_ids: List[str],
    ) -> bool:
        """
        Adds the specified job_ids to the specified instance's running_jobs. This will
        allocate resources_allocated_per_job for each job from the instance's
        available_resources.

        This function should return True if successful, False if we fail an optimistic
        concurrency check OR if the job_id is already in use by the specified instance,
        and raise an exception for any other issues.

        TODO consider more robust handling of job_id collisions--probably indicates a
        consistency issue
        """
        pass

    def allocate_jobs_to_instance_max_chunk(self) -> int:
        """
        allocate_jobs_to_instance in some cases has relatively low limits for how many
        jobs can be allocated to an instance a time. This tells the caller that
        new_job_ids in allocate_jobs_to_instance shouldn't be larger than the limit
        returned by this function.
        """
        return -1

    @abc.abstractmethod
    async def deallocate_job_from_instance(
        self, instance: _TInstanceState, job_id: str
    ) -> bool:
        """
        Removes the specified job from the specified instance and restores the resources
        that were allocated to that job.

        Returns True if the job was removed, returns False if the job does not exist
        (e.g. it was already removed or never existed in the first place).
        """
        pass

    @abc.abstractmethod
    async def launch_instances(
        self,
        resources_required_per_task: ResourcesInternal,
        num_concurrent_tasks: int,
        alloc_cloud_instances: AllocVM,
        abort: Optional[asyncio.Event],
    ) -> Sequence[CloudInstance]:
        """
        This isn't per se part of the "instance registration" process, but it's helpful
        to have on the same object.
        """
        pass

    @abc.abstractmethod
    async def set_prevent_further_allocation(self, name: str, value: bool) -> bool:
        pass

    @abc.abstractmethod
    async def authorize_current_ip(self, alloc_cloud_instances: AllocVM) -> None:
        pass

    @abc.abstractmethod
    async def open_ports(
        self,
        ports: Optional[Sequence[str]],
        allocated_existing_instances: Iterable[_TInstanceState],
        allocated_new_instances: Iterable[CloudInstance],
    ) -> None:
        pass


@dataclasses.dataclass
class _InstanceWithProposedJobs(Generic[_TInstanceState]):
    """Just used in _choose_existing_instances"""

    orig_instance: _TInstanceState
    proposed_jobs: List[str]
    proposed_available_resources: ResourcesInternal


async def _choose_existing_instances(
    instance_registrar: InstanceRegistrar[_TInstanceState],
    resources_required_per_job: ResourcesInternal,
    num_jobs: int,
) -> Tuple[Dict[Tuple[str, str], List[str]], Dict[Tuple[str, str], _TInstanceState]]:
    """
    Chooses existing registered instances to run the specified job(s). The general
    strategy is to pack instances as tightly as possible to allow larger jobs to come
    along later.

    Returns two dictionaries: {(public_address, name): [job_ids]}, {(public_address,
    name): instance_state}. The keys of the two dictionaries will be the same set of
    public addresses.
    """

    # these represent jobs that have been allocated in the InstanceRegistrar
    num_jobs_allocated = 0
    # {public_address: [job_ids]}
    allocated_jobs: Dict[Tuple[str, str], List[str]] = {}
    # {public_address: instance_state}
    allocated_instances: Dict[Tuple[str, str], _TInstanceState] = {}

    # try to allocate a maximum of 3 times. We will retry if there's an optimistic
    # concurrency issue (i.e. someone else allocates to an instance at the same time
    # as us). Failing just means that we're unable to allocate the jobs.
    i = 0
    all_success = False
    while i < 3 and not all_success:
        instances = [
            _InstanceWithProposedJobs(instance, [], instance.get_available_resources())
            for instance in await instance_registrar.get_registered_instances()
            if not instance.prevent_further_allocation
        ]

        sort_keys = [
            remaining_resources_sort_key(
                instance.proposed_available_resources, resources_required_per_job
            )
            for instance in instances
        ]

        # these represent proposed allocations--they are not actually allocated
        # until we update InstanceRegistrar
        num_jobs_proposed = 0

        if sort_keys:
            while num_jobs_allocated + num_jobs_proposed < num_jobs:
                # choose an instance
                chosen_index = min(range(len(sort_keys)), key=lambda i: sort_keys[i])
                # if the indicator is 1, that means none of the instances can run our
                # job
                if sort_keys[chosen_index][0] == 1:
                    break

                # we successfully chose an instance!
                chosen_instance = instances[chosen_index]
                chosen_instance.proposed_jobs.append(str(uuid.uuid4()))
                num_jobs_proposed += 1

                # decrease the agent's available_resources
                chosen_instance.proposed_available_resources = assert_is_not_none(
                    (
                        chosen_instance.proposed_available_resources.subtract(
                            resources_required_per_job
                        )
                    )
                )
                # decrease the sort key for the chosen agent
                sort_keys[chosen_index] = remaining_resources_sort_key(
                    chosen_instance.proposed_available_resources,
                    resources_required_per_job,
                )

        # now that we've chosen which instance(s) will run our job(s), try to actually
        # get the allocation in the InstanceRegistrar. This could fail if another
        # process is trying to do an allocation at the same time as us so the instances
        # we've chosen actually don't have enough resources (even though they did at the
        # top of this function).

        allocate_jobs_max_chunk = (
            instance_registrar.allocate_jobs_to_instance_max_chunk()
        )
        all_success = True
        for top_instance in instances:
            if top_instance.proposed_jobs:
                if allocate_jobs_max_chunk != -1:
                    # split proposed_jobs into chunks of size allocate_jobs_max_chunk
                    it = iter(top_instance.proposed_jobs)
                    chunked_instances: Iterable[_InstanceWithProposedJobs] = (
                        dataclasses.replace(
                            top_instance, proposed_jobs=proposed_jobs_chunk
                        )
                        for proposed_jobs_chunk in iter(
                            lambda: list(itertools.islice(it, allocate_jobs_max_chunk)),
                            [],
                        )
                    )
                else:
                    chunked_instances = (top_instance,)

                for instance in chunked_instances:
                    success = await instance_registrar.allocate_jobs_to_instance(
                        # it's very important to pass the orig_instance
                        # here--allocate_jobs_to_instance can rely on
                        # available_resources being correct and we want the original
                        # one, not the proposed_available_resources that we've modified.
                        instance.orig_instance,
                        resources_required_per_job,
                        instance.proposed_jobs,
                    )
                    if success:
                        allocated_jobs.setdefault(
                            (
                                instance.orig_instance.public_address,
                                instance.orig_instance.name,
                            ),
                            [],
                        ).extend(instance.proposed_jobs)
                        allocated_instances[
                            (
                                instance.orig_instance.public_address,
                                instance.orig_instance.name,
                            )
                        ] = instance.orig_instance
                        num_jobs_allocated += len(instance.proposed_jobs)
                    else:
                        all_success = False

        i += 1

    if num_jobs == 1:
        if num_jobs_allocated == 0:
            print(
                "Job was not allocated to any existing instances, will launch a new"
                " instance"
            )
        else:
            print(
                "Job was allocated to an existing instances: "
                + " ".join(
                    f"{public_address} ({name})"
                    for public_address, name in allocated_jobs.keys()
                )
            )
    else:
        print(
            f"{num_jobs_allocated}/{num_jobs} workers allocated to existing instances: "
            + " ".join(
                f"{public_address} ({name})"
                for public_address, name in allocated_jobs.keys()
            )
        )

    return allocated_jobs, allocated_instances


async def _launch_new_instances(
    instance_registrar: InstanceRegistrar,
    instance_type_resources_required_per_task: ResourcesInternal,
    num_concurrent_tasks: int,
    alloc_cloud_instance: AllocVM,
    original_num_concurrent_tasks: int,
    abort: Optional[asyncio.Event],
) -> Tuple[Dict[Tuple[str, str], List[str]], Dict[Tuple[str, str], CloudInstance]]:
    """
    Chooses the cheapest instances to launch that can run the specified jobs, launches
    them, adds them to the InstanceRegistrar, and allocates the specified jobs to them.

    Returns two dictionaries, {(public_address, name): [job_ids]}, {(public_address,
    name): cloud_instance}. The keys of the two dictionaries will be the same.

    The CloudInstance objects returned will have a resources field that includes both
    the instance type resources (like CPU and memory) that were used to select an
    instance type as well as the "runtime resources" provided by alloc_cloud_instance
    (like the AMI id and subnet ID) which describe how the instance was actaully
    launched.

    original_num_jobs is only needed to produce more coherent logging.
    """

    instances = await instance_registrar.launch_instances(
        instance_type_resources_required_per_task,
        num_concurrent_tasks,
        alloc_cloud_instance,
        abort,
    )

    description_strings = []
    total_num_allocated_jobs = 0
    total_cost_per_hour: float = 0
    allocated_jobs = {}
    allocated_instances = {}

    for instance in instances:
        # just to make the code more readable
        instance_info = instance.instance_type.instance_type
        all_resources = instance_info.resources.combine(
            alloc_cloud_instance.get_runtime_resources()
        )

        # the number of jobs to allocate to this instance
        num_allocated_jobs = min(
            num_concurrent_tasks - total_num_allocated_jobs,
            instance.instance_type.workers_per_instance_full,
        )
        total_num_allocated_jobs += num_allocated_jobs
        job_ids = [str(uuid.uuid4()) for _ in range(num_allocated_jobs)]

        await instance_registrar.register_instance(
            instance.public_dns_name,
            instance.name,
            assert_is_not_none(
                all_resources.subtract(
                    instance_type_resources_required_per_task.multiply(
                        num_allocated_jobs
                    )
                )
            ),
            [(job_id, instance_type_resources_required_per_task) for job_id in job_ids],
        )

        allocated_jobs[(instance.public_dns_name, instance.name)] = job_ids
        allocated_instances[(instance.public_dns_name, instance.name)] = instance
        description_strings.append(
            f"{instance.public_dns_name}: {instance_info.name} "
            f"({instance_info.resources.format_cpu_memory_gpu()}), "
            f"{instance_info.on_demand_or_spot} (${instance_info.price}/hr, "
            f"{instance_info.resources.format_eviction_rate()}), will run "
            f"{num_allocated_jobs} workers"
        )
        total_cost_per_hour += instance_info.price

    if original_num_concurrent_tasks == 1:
        # there should only ever be one description_strings
        print(f"Launched a new instance for the job: {' '.join(description_strings)}")
    else:
        print(
            f"Launched {len(description_strings)} new instance(s) (total "
            f"${total_cost_per_hour}/hr) for the remaining "
            f"{num_concurrent_tasks} workers:\n"
            + "\n".join(["\t" + s for s in description_strings])
        )

    return allocated_jobs, allocated_instances


async def allocate_jobs_to_instances(
    instance_registrar: InstanceRegistrar,
    resources_required_per_task: ResourcesInternal,
    num_concurrent_tasks: int,
    alloc_cloud_instance: AllocVM,
    ports: Optional[Sequence[str]],
    abort: Optional[asyncio.Event],
) -> AsyncIterable[Dict[Tuple[str, str], List[str]]]:
    """
    This function first tries to re-use existing instances, and if necessary launches
    the cheapest possible new instances that have the requested resources.

    Returns {(public_address, name): [job_ids]}
    """

    authorize_current_ip_task = asyncio.create_task(
        instance_registrar.authorize_current_ip(alloc_cloud_instance)
    )

    (
        existing_instances_allocated_jobs,
        allocated_existing_instances,
    ) = await _choose_existing_instances(
        instance_registrar,
        resources_required_per_task.combine(
            alloc_cloud_instance.get_runtime_resources()
        ),
        num_concurrent_tasks,
    )

    if len(existing_instances_allocated_jobs) > 0:
        await instance_registrar.open_ports(
            ports, allocated_existing_instances.values(), ()
        )
        await authorize_current_ip_task  # this should almost always be complete by now
        yield existing_instances_allocated_jobs

    num_concurrent_tasks_remaining = num_concurrent_tasks - sum(
        len(jobs) for jobs in existing_instances_allocated_jobs.values()
    )

    if num_concurrent_tasks_remaining > 0:
        (
            new_instances_allocated_jobs,
            allocated_new_instances,
        ) = await _launch_new_instances(
            instance_registrar,
            resources_required_per_task,
            num_concurrent_tasks_remaining,
            alloc_cloud_instance,
            num_concurrent_tasks,
            abort,
        )
        # TODO if this fails and we have existing instances, we should just carry on
        # with those

        await instance_registrar.open_ports(ports, (), allocated_new_instances.values())
        if len(existing_instances_allocated_jobs) == 0:
            # we need to await this if and only if we didn't await it before
            await authorize_current_ip_task

        yield new_instances_allocated_jobs


async def allocate_single_job_to_instance(
    instance_registrar: InstanceRegistrar,
    resources_required: ResourcesInternal,
    alloc_cloud_instance: AllocVM,
    ports: Optional[Sequence[str]],
) -> Tuple[str, str]:
    """
    A thin wrapper around allocate_jobs_to_instances with a simpler return type for when
    you only need one job.
    """

    result = None

    async for allocated_hosts in allocate_jobs_to_instances(
        instance_registrar, resources_required, 1, alloc_cloud_instance, ports, None
    ):
        if len(allocated_hosts) == 0:
            pass
        elif len(allocated_hosts) > 1:
            raise ValueError(f"Requested 1 host, but got back {len(allocated_hosts)}")
        else:
            if result is None:
                (public_address, name), job_ids = list(allocated_hosts.items())[0]
                if len(job_ids) != 1:
                    raise ValueError(
                        f"Requested 1 job allocation but got back {len(job_ids)}"
                    )
                result = public_address, job_ids[0]
            else:
                raise ValueError("Requested 1 host, but got back more than one")

    if result is None:
        raise ValueError("Requested 1 host, but did not get any back")

    return result



import glob
import itertools
import os
import sys
import uuid
import zipfile
from os.path import join, realpath, splitext
from typing import Iterable, List, Set, Tuple

from meadowrun.shared import create_zipfile


def zip_local_code(
    result_zip_dir: str,
    include_sys_path: bool = True,
    additional_python_paths: Iterable[str] = tuple(),
    python_paths_extensions: Iterable[str] = (".py",),
    working_directory_globs: Iterable[str] = tuple(),
) -> Tuple[str, List[str], str]:
    """
    The goal of this function is to zip up "the current local python code excluding the
    interpreter code" as well as any specified files in the current working directory.

    There are two types of folders/files that make it into the zip file. The first are
    "python paths". Assuming include_sys_path is True, "python paths" will be
    additional_python_paths + paths in sys.path that are NOT part of the interpreter,
    i.e. paths that do NOT start with sys.prefix, etc. The second type are files
    specified by working_directory_globs which are always relative to the current
    working directory.

    python paths will only include *.py files (or whatever is specified in
    python_paths_extensions).

    Args:
        result_zip_dir: directory where to put the resulting zip file.
        include_sys_path: Whether to zip paths on sys.path. Paths that are also in
            sys.prefix or sys.base_prefix are automatically excluded. Defaults to True.
        additional_python_paths: Additional python paths to zip. Defaults to empty
            tuple, i.e. no additional python paths.
        python_paths_extensions: Only files with these extensions are zipped. Defaults
            to (".py", ".pyd"), i.e. only python files.
        working_directory_globs: Globs that will be passed directly to glob.glob to
            specify any file relative to the current working directory. Globs will only
            match files.

    Raises:
        ValueError: If there are no paths to zip.

    Returns:
        Tuple[str, List[str], str]: path to zip file in result_zip_dir, python paths in
        the zip file, current working directory in the zip file
    """
    current_working_directory = realpath(os.getcwd())
    non_python_paths_to_zip = [current_working_directory]
    current_working_directory_parent = os.path.dirname(current_working_directory)
    # this will hold (path_to_file, path_to_containing_dir) for every file that matches
    # a glob
    real_file_paths_to_zip = []
    for file_glob in working_directory_globs:
        for file in glob.glob(file_glob, recursive=True):
            if os.path.isfile(file):
                file_real_path = realpath(file)
                if not file_real_path.startswith(current_working_directory_parent):
                    raise ValueError(
                        "Specifying files outside of the current working directory's "
                        "parent is not supported. I.e. ../foo is fine, but ../../foo "
                        "will not work"
                    )
                dir_real_path = os.path.dirname(file_real_path)
                real_file_paths_to_zip.append((file_real_path, dir_real_path))
                non_python_paths_to_zip.append(dir_real_path)

    real_python_paths_to_zip = _get_python_paths_to_zip(
        additional_python_paths, include_sys_path
    )

    (
        python_zip_paths,
        non_python_zip_paths,
        real_python_root_paths_to_zip_paths,
    ) = _consolidate_paths_to_zip(real_python_paths_to_zip, non_python_paths_to_zip)

    # Currently we always zip all the files and then compute the hash. The hash is just
    # used to avoid upload to S3. In future we could try to split the files into chunks
    # or even individually, to only upload what we need. For small-ish codebases zipping
    # and hashing is fast though, certainly relative to EC2 startup times and container
    # building, so this is not currently a bottleneck.
    zip_file_path = join(result_zip_dir, str(uuid.uuid4()) + ".zip")
    # ZIP_DEFLATED because it's the fastest. All code gets zipped every time, even when
    # it hasn't changed, so this needs to be fast above all else.
    with create_zipfile(zip_file_path, "w", zipfile.ZIP_DEFLATED) as zip_file:
        for real_path, zip_path in real_python_root_paths_to_zip_paths:
            for dirpath, _, filenames in os.walk(real_path):
                for filename in filenames:
                    if splitext(filename)[1].lower() in python_paths_extensions:
                        full = join(dirpath, filename)
                        full_zip = full.replace(real_path, zip_path, 1)
                        zip_file.write(full, full_zip)

        for (file_real_path, dir_real_path), dir_zip_path in zip(
            real_file_paths_to_zip, itertools.islice(non_python_zip_paths, 1, None)
        ):
            file_zip_path = file_real_path.replace(dir_real_path, dir_zip_path, 1)
            if os.path.sep != "/":
                # we have to make the sub-paths compatible with Linux which is our
                # target OS
                file_zip_path = file_zip_path.replace(os.path.sep, "/")
            zip_file.write(file_real_path, file_zip_path)

    return zip_file_path, python_zip_paths, non_python_zip_paths[0]


def _get_python_paths_to_zip(
    python_paths_to_zip: Iterable[str],
    include_sys_path: bool,
) -> List[str]:
    python_paths_to_zip = [realpath(path) for path in python_paths_to_zip]
    if include_sys_path:
        for candidate in sys.path:
            if not (
                candidate.startswith(sys.prefix)
                or candidate.startswith(sys.base_prefix)
            ):
                if candidate == "":
                    # empty path on sys.path means cwd
                    python_paths_to_zip.append(realpath(os.getcwd()))
                else:
                    python_paths_to_zip.append(realpath(candidate))
    return python_paths_to_zip


def _consolidate_paths_to_zip(
    real_python_paths_to_zip: Iterable[str], real_non_python_paths_to_zip: Iterable[str]
) -> Tuple[List[str], List[str], List[Tuple[str, str]]]:
    """
    Takes some real paths, and returns the path they should be in the zip. Some example
    inputs/outputs of real paths to zip paths:
    - [a/b/c, a/b] -> [b/c, b] (don't make c a separate directory because it's already
      included in b
    - [a/b, d/b] ->  [b, b_0] (we have two b folders that we need to keep separate)

    Return values are python_zip_paths, non_python_zip_paths, python_roots

    python_zip_paths and non_python_zip paths correspond 1-to-1 with the elements in
    real_python_paths_to_zip and real_non_python_paths_to_zip.

    python_roots is a subset of zip(real_python_paths_to_zip, python_zip_paths) that are
    "roots". E.g. in the first example above, this will just return [(a/b, b)].
    """

    # This whole rigmarole because:
    # 1. Any folders that are subfolders of folders already zipped needn't be zipped
    # 2. Need unique top-level folder name in zip file, that isn't full path
    # 3. each path in real_paths_to_zip must map to a folder in zip file.

    # remembers and updates used top-level folder name in the zip, and generates unique
    # names (keeping the original as much as possible)
    def find_unused_path(used_paths: Set[str], path_base: str) -> str:
        candidate = path_base
        for i in range(100):
            if candidate not in used_paths:
                used_paths.add(candidate)
                return candidate
            candidate = f"{path_base}_{i}"
        raise Exception(f"Gave up after 100 paths with base {path_base}")

    # sorting so shortest (i.e. superset) paths come first
    root_candidates = tuple(
        sorted(itertools.chain(real_python_paths_to_zip, real_non_python_paths_to_zip))
    )

    current_root = root_candidates[0]
    zip_path = os.path.basename(current_root)
    real_paths_to_zip_paths = {current_root: zip_path}
    real_root_paths = [current_root]
    used_zip_paths = {zip_path}

    for candidate in root_candidates:
        if candidate.startswith(current_root):
            # we have a subpath. Record it.
            zip_path = candidate.replace(
                current_root, real_paths_to_zip_paths[current_root], 1
            )
            if os.path.sep != "/":
                # we have to make the sub-paths compatible with Linux which is our
                # target OS
                zip_path = zip_path.replace(os.path.sep, "/")
            real_paths_to_zip_paths[candidate] = zip_path
        else:
            # we have a new root.
            current_root = candidate
            # give it a unique name in the zip.
            zip_path = find_unused_path(used_zip_paths, os.path.basename(current_root))
            real_paths_to_zip_paths[current_root] = zip_path
            real_root_paths.append(current_root)

    # same order as input paths, but paths in the zip
    zip_python_paths = [
        real_paths_to_zip_paths[real_path] for real_path in real_python_paths_to_zip
    ]
    zip_non_python_paths = [
        real_paths_to_zip_paths[real_path] for real_path in real_non_python_paths_to_zip
    ]

    # consolidate python paths
    python_root_candidates = tuple(sorted(real_python_paths_to_zip))
    current_root = python_root_candidates[0]
    real_python_root_paths = [current_root]
    for candidate in python_root_candidates:
        if not candidate.startswith(current_root):
            # we have a new root.
            current_root = candidate
            real_python_root_paths.append(current_root)

    return (
        zip_python_paths,
        zip_non_python_paths,
        [
            (real_path, real_paths_to_zip_paths[real_path])
            for real_path in real_python_root_paths
        ],
    )



from __future__ import annotations

import asyncio
from functools import lru_cache
from typing import Any, Callable, Iterable, Optional, Set, TypeVar, Union

import aiohttp
import aiohttp.client_exceptions
import boto3
import botocore.exceptions

_T = TypeVar("_T")

# the IAM user group that gives users the permissions they need to run Meadowrun jobs
_MEADOWRUN_USER_GROUP_NAME = "meadowrun_user_group"


class MeadowrunNotInstalledError(Exception):
    def __init__(self, missing_resource: str):
        super().__init__(
            f"Meadowrun resource is not available: {missing_resource}. Please run "
            "`meadowrun-manage-ec2 install` as an AWS root/Administrator account"
        )


class MeadowrunAWSAccessError(Exception):
    def __init__(self, resource: str, custom_exception_message: bool = False):
        if custom_exception_message:
            message = resource
        else:
            message = (
                f"Encountered an AWS access exception trying to access {resource}. If "
                f"you are not a member of the {_MEADOWRUN_USER_GROUP_NAME} user group, "
                "please add yourself/ask to to be added to this group with `aws iam "
                f"add-user-to-group --group-name {_MEADOWRUN_USER_GROUP_NAME} "
                "--user-name <username>`. If this group does not exist, please run "
                "`meadowrun-manage-ec2 install` as an AWS root/Administrator account"
            )
        super().__init__(message)


def wrap_access_or_install_errors(
    func: Callable[[], _T],
    resource: str,
    access_error_code: Union[str, Set[str]],
    install_error_code: Union[str, Set[str]],
) -> _T:
    """
    A variation on ignore_boto3_error_code that would otherwise require wrapping func in
    two lambdas. Makes sense for operations that are the first operation a user will
    try, so that we can give them helpful information if Meadowrun hasn't been installed
    yet or they don't have the right permissions.
    """
    if isinstance(access_error_code, str):
        access_error_code = {access_error_code}

    if isinstance(install_error_code, str):
        install_error_code = {install_error_code}

    try:
        return func()
    except botocore.exceptions.ClientError as e:
        if "Error" in e.response:
            error = e.response["Error"]
            if "Code" in error:
                if error["Code"] in access_error_code:
                    raise MeadowrunAWSAccessError(resource) from e
                elif error["Code"] in install_error_code:
                    raise MeadowrunNotInstalledError(resource) from e

        raise


def _boto3_paginate(method: Any, **kwargs: Any) -> Iterable[Any]:
    paginator = method.__self__.get_paginator(method.__name__)
    for page in paginator.paginate(**kwargs).result_key_iters():
        for item in page:
            yield item


# on WSL, the EC2 metadata endpoint hangs instead of immediately failing to connect. We
# have a relatively short timeout here as we should never have network issues with the
# EC2 metadata endpoint
_EC2_METADATA_ENDPOINT_TIMEOUT_SECS = 1


async def _get_ec2_metadata(url_suffix: str) -> Optional[str]:
    """
    Queries the EC2 metadata endpoint, which gives us information about the EC2 instance
    we're currently running on:
    https://docs.aws.amazon.com/AWSEC2/latest/UserGuide/instancedata-data-retrieval.html
    Returns None if the endpoint is not available because e.g. we're not running on an
    EC2 instance.
    """
    try:
        async with aiohttp.request(
            "GET",
            f"http://169.254.169.254/latest/meta-data/{url_suffix}",
            timeout=aiohttp.ClientTimeout(total=_EC2_METADATA_ENDPOINT_TIMEOUT_SECS),
        ) as response:
            response.raise_for_status()
            return await response.text()
    except asyncio.CancelledError:
        raise
    except Exception:
        # the AWS metadata endpoint is not available, probably because we're not on
        # an EC2 instance.
        pass

    return None


async def _get_default_region_name() -> str:
    """
    Tries to get the default region name. E.g. us-east-2. First sees if the AWS CLI is
    set up, and returns the equivalent of `aws configure get region`. Then checks if we
    are running on an EC2 instance in which case we check the AWS metadata endpoint

    TODO this function is overused almost everywhere. Currently, we just always use the
    default region for everything and don't support multi-region deployments, but region
    should be a first-class concept.
    """

    default_session = boto3._get_default_session()
    if default_session is not None and default_session.region_name:
        # equivalent of `aws configure get region`
        return default_session.region_name
    else:
        result = await _get_ec2_metadata("placement/region")
        if result:
            return result
        else:
            raise ValueError(
                "region_name was not specified, we are not logged into AWS locally, and"
                " we're not running on an EC2 instance, so we have no way of picking a "
                "default region."
            )


def _iam_role_exists(iam_client: Any, role_name: str) -> bool:
    """Returns True if the specified IAM role exists, otherwise returns False"""
    try:
        iam_client.get_role(RoleName=role_name)
        return True
    except Exception as e:
        # Unfortunately boto3 appears to have dynamic exception types. So type(e) would
        # be "botocore.errorfactory.NoSuchEntityException", but NoSuchEntityException
        # can't be imported from botocore.errorfactory.
        if type(e).__name__ == "NoSuchEntityException":
            return False

        raise


async def _get_current_ip_for_ssh() -> str:
    """
    Gets an ip address for the current machine that is likely to work for allowing SSH
    into an EC2 instance.
    """
    # if we're already in an EC2 instance, use the EC2 metadata to get our private IP
    private_ip = await _get_ec2_metadata("local-ipv4")
    if private_ip:
        return private_ip

    # otherwise, we'll use checkip.amazonaws.com to figure out how AWS sees our IP
    async with aiohttp.request("GET", "https://checkip.amazonaws.com/") as response:
        response.raise_for_status()
        return (await response.text()).strip()


@lru_cache(maxsize=1)
def _get_account_number() -> str:
    # weird that we have to do this to get the account number to construct the ARN
    return boto3.client("sts").get_caller_identity().get("Account")


BUCKET_PREFIX = "meadowrun"


def get_bucket_name(region_name: str) -> str:
    # s3 bucket names must be globally unique across all accounts and regions.
    return f"{BUCKET_PREFIX}-{region_name}-{_get_account_number()}"



"""
Written to work with Kubernetes readiness probes. Returns 0 if there are no jobs running
(i.e. we are ready), 1 if there is a job running (we are not ready for another job)
"""

import sys

import filelock


_JOB_IS_RUNNING_FILE = "/var/meadowrun/job_is_running"


def is_job_running() -> bool:
    try:
        with filelock.FileLock(_JOB_IS_RUNNING_FILE, 0):
            pass
        return False
    except filelock.Timeout:
        return True


def main() -> None:
    if is_job_running():
        sys.exit(1)
    else:
        sys.exit(0)


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: meadowrun/meadowrun.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x19meadowrun/meadowrun.proto\x12\tmeadowrun"(\n\nStringPair\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t"+\n\x15ServerAvailableFolder\x12\x12\n\ncode_paths\x18\x01 \x03(\t"@\n\x0b\x43odeZipFile\x12\x0b\n\x03url\x18\x01 \x01(\t\x12\x12\n\ncode_paths\x18\x02 \x03(\t\x12\x10\n\x08\x63wd_path\x18\x03 \x01(\t"I\n\rGitRepoCommit\x12\x10\n\x08repo_url\x18\x01 \x01(\t\x12\x0e\n\x06\x63ommit\x18\x02 \x01(\t\x12\x16\n\x0epath_to_source\x18\x03 \x01(\t"I\n\rGitRepoBranch\x12\x10\n\x08repo_url\x18\x01 \x01(\t\x12\x0e\n\x06\x62ranch\x18\x02 \x01(\t\x12\x16\n\x0epath_to_source\x18\x03 \x01(\t"6\n\x1aServerAvailableInterpreter\x12\x18\n\x10interpreter_path\x18\x01 \x01(\t"7\n\x11\x43ontainerAtDigest\x12\x12\n\nrepository\x18\x01 \x01(\t\x12\x0e\n\x06\x64igest\x18\x02 \x01(\t"1\n\x0e\x43ontainerAtTag\x12\x12\n\nrepository\x18\x01 \x01(\t\x12\x0b\n\x03tag\x18\x02 \x01(\t"\x8d\x02\n\x15\x45nvironmentSpecInCode\x12\x34\n\x10\x65nvironment_type\x18\x01 \x01(\x0e\x32\x1a.meadowrun.EnvironmentType\x12\x14\n\x0cpath_to_spec\x18\x02 \x01(\t\x12\x16\n\x0epython_version\x18\x03 \x01(\t\x12U\n\x13\x61\x64\x64itional_software\x18\x05 \x03(\x0b\x32\x38.meadowrun.EnvironmentSpecInCode.AdditionalSoftwareEntry\x1a\x39\n\x17\x41\x64\x64itionalSoftwareEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01"\x8c\x02\n\x0f\x45nvironmentSpec\x12\x34\n\x10\x65nvironment_type\x18\x01 \x01(\x0e\x32\x1a.meadowrun.EnvironmentType\x12\x0c\n\x04spec\x18\x02 \x01(\t\x12\x11\n\tspec_lock\x18\x03 \x01(\t\x12\x16\n\x0epython_version\x18\x04 \x01(\t\x12O\n\x13\x61\x64\x64itional_software\x18\x05 \x03(\x0b\x32\x32.meadowrun.EnvironmentSpec.AdditionalSoftwareEntry\x1a\x39\n\x17\x41\x64\x64itionalSoftwareEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01".\n\x18ServerAvailableContainer\x12\x12\n\nimage_name\x18\x01 \x01(\t"G\n\x0cPyCommandJob\x12\x14\n\x0c\x63ommand_line\x18\x01 \x03(\t\x12!\n\x19pickled_context_variables\x18\x02 \x01(\x0c"C\n\x15QualifiedFunctionName\x12\x13\n\x0bmodule_name\x18\x01 \x01(\t\x12\x15\n\rfunction_name\x18\x02 \x01(\t"\xa5\x01\n\rPyFunctionJob\x12\x43\n\x17qualified_function_name\x18\x01 \x01(\x0b\x32 .meadowrun.QualifiedFunctionNameH\x00\x12\x1a\n\x10pickled_function\x18\x02 \x01(\x0cH\x00\x12"\n\x1apickled_function_arguments\x18\x03 \x01(\x0c\x42\x0f\n\rfunction_spec"\xc8\x01\n\nPyAgentJob\x12\x43\n\x17qualified_function_name\x18\x01 \x01(\x0b\x32 .meadowrun.QualifiedFunctionNameH\x00\x12\x1a\n\x10pickled_function\x18\x02 \x01(\x0cH\x00\x12\x1e\n\x16pickled_agent_function\x18\x03 \x01(\x0c\x12(\n pickled_agent_function_arguments\x18\x04 \x01(\x0c\x42\x0f\n\rfunction_spec"P\n\x08GridTask\x12\x0f\n\x07task_id\x18\x01 \x01(\x05\x12\x0f\n\x07\x61ttempt\x18\x03 \x01(\x05\x12"\n\x1apickled_function_arguments\x18\x02 \x01(\x0c"\xf4\x01\n\x0e\x43ontainerImage\x12\x41\n\x19\x63ontainer_image_at_digest\x18\x01 \x01(\x0b\x32\x1c.meadowrun.ContainerAtDigestH\x00\x12;\n\x16\x63ontainer_image_at_tag\x18\x02 \x01(\x0b\x32\x19.meadowrun.ContainerAtTagH\x00\x12O\n server_available_container_image\x18\x03 \x01(\x0b\x32#.meadowrun.ServerAvailableContainerH\x00\x42\x11\n\x0f\x63ontainer_image"\xd8\x08\n\x03Job\x12\x0e\n\x06job_id\x18\x01 \x01(\t\x12\x19\n\x11job_friendly_name\x18\x02 \x01(\t\x12\x43\n\x17server_available_folder\x18\x05 \x01(\x0b\x32 .meadowrun.ServerAvailableFolderH\x00\x12\x33\n\x0fgit_repo_commit\x18\x06 \x01(\x0b\x32\x18.meadowrun.GitRepoCommitH\x00\x12\x33\n\x0fgit_repo_branch\x18\x07 \x01(\x0b\x32\x18.meadowrun.GitRepoBranchH\x00\x12/\n\rcode_zip_file\x18\x13 \x01(\x0b\x32\x16.meadowrun.CodeZipFileH\x00\x12M\n\x1cserver_available_interpreter\x18\x08 \x01(\x0b\x32%.meadowrun.ServerAvailableInterpreterH\x01\x12;\n\x13\x63ontainer_at_digest\x18\t \x01(\x0b\x32\x1c.meadowrun.ContainerAtDigestH\x01\x12\x35\n\x10\x63ontainer_at_tag\x18\n \x01(\x0b\x32\x19.meadowrun.ContainerAtTagH\x01\x12I\n\x1aserver_available_container\x18\x0b \x01(\x0b\x32#.meadowrun.ServerAvailableContainerH\x01\x12\x44\n\x18\x65nvironment_spec_in_code\x18\x0c \x01(\x0b\x32 .meadowrun.EnvironmentSpecInCodeH\x01\x12\x36\n\x10\x65nvironment_spec\x18\x12 \x01(\x0b\x32\x1a.meadowrun.EnvironmentSpecH\x01\x12\x35\n\x12sidecar_containers\x18\x16 \x03(\x0b\x32\x19.meadowrun.ContainerImage\x12\x34\n\x15\x65nvironment_variables\x18\r \x03(\x0b\x32\x15.meadowrun.StringPair\x12&\n\x1eresult_highest_pickle_protocol\x18\x0e \x01(\x05\x12-\n\npy_command\x18\x0f \x01(\x0b\x32\x17.meadowrun.PyCommandJobH\x02\x12/\n\x0bpy_function\x18\x10 \x01(\x0b\x32\x18.meadowrun.PyFunctionJobH\x02\x12)\n\x08py_agent\x18\x17 \x01(\x0b\x32\x15.meadowrun.PyAgentJobH\x02\x12@\n\x13\x63redentials_sources\x18\x11 \x03(\x0b\x32#.meadowrun.CredentialsSourceMessage\x12\r\n\x05ports\x18\x14 \x03(\t\x12\x10\n\x08uses_gpu\x18\x15 \x01(\x08\x42\x11\n\x0f\x63ode_deploymentB\x18\n\x16interpreter_deploymentB\n\n\x08job_spec"\xc2\x03\n\x0cProcessState\x12\x37\n\x05state\x18\x01 \x01(\x0e\x32(.meadowrun.ProcessState.ProcessStateEnum\x12\x0b\n\x03pid\x18\x02 \x01(\x05\x12\x14\n\x0c\x63ontainer_id\x18\x03 \x01(\t\x12\x15\n\rlog_file_name\x18\x04 \x01(\t\x12\x16\n\x0epickled_result\x18\x05 \x01(\x0c\x12\x13\n\x0breturn_code\x18\x06 \x01(\x05\x12\x1a\n\x12max_memory_used_gb\x18\x07 \x01(\x02"\xf5\x01\n\x10ProcessStateEnum\x12\x0b\n\x07\x44\x45\x46\x41ULT\x10\x00\x12\x11\n\rRUN_REQUESTED\x10\x01\x12\x0b\n\x07RUNNING\x10\x02\x12\r\n\tSUCCEEDED\x10\x03\x12\x16\n\x12RUN_REQUEST_FAILED\x10\x04\x12\x14\n\x10PYTHON_EXCEPTION\x10\x05\x12\x18\n\x14NON_ZERO_RETURN_CODE\x10\x06\x12\x1b\n\x17RESOURCES_NOT_AVAILABLE\x10\x07\x12\x17\n\x13\x45RROR_GETTING_STATE\x10\x08\x12\x1a\n\x16UNEXPECTED_WORKER_EXIT\x10\n\x12\x0b\n\x07UNKNOWN\x10\t"P\n\x0eJobStateUpdate\x12\x0e\n\x06job_id\x18\x01 \x01(\t\x12.\n\rprocess_state\x18\x02 \x01(\x0b\x32\x17.meadowrun.ProcessState"i\n\x15GridTaskStateResponse\x12\x0f\n\x07task_id\x18\x01 \x01(\x05\x12\x0f\n\x07\x61ttempt\x18\x03 \x01(\x05\x12.\n\rprocess_state\x18\x02 \x01(\x0b\x32\x17.meadowrun.ProcessState"\xd0\x02\n\x18\x43redentialsSourceMessage\x12/\n\x07service\x18\x01 \x01(\x0e\x32\x1e.meadowrun.Credentials.Service\x12\x13\n\x0bservice_url\x18\x02 \x01(\t\x12/\n\naws_secret\x18\x03 \x01(\x0b\x32\x19.meadowrun.AwsSecretProtoH\x00\x12\x33\n\x0c\x61zure_secret\x18\x05 \x01(\x0b\x32\x1b.meadowrun.AzureSecretProtoH\x00\x12?\n\x15server_available_file\x18\x04 \x01(\x0b\x32\x1e.meadowrun.ServerAvailableFileH\x00\x12=\n\x11kubernetes_secret\x18\x06 \x01(\x0b\x32 .meadowrun.KubernetesSecretProtoH\x00\x42\x08\n\x06source"\x95\x01\n\x0b\x43redentials\x12\x13\n\x0b\x63redentials\x18\x01 \x01(\x0c"3\n\x07Service\x12\x13\n\x0f\x44\x45\x46\x41ULT_SERVICE\x10\x00\x12\n\n\x06\x44OCKER\x10\x01\x12\x07\n\x03GIT\x10\x02"<\n\x04Type\x12\x10\n\x0c\x44\x45\x46\x41ULT_TYPE\x10\x00\x12\x15\n\x11USERNAME_PASSWORD\x10\x01\x12\x0b\n\x07SSH_KEY\x10\x02"\\\n\x0e\x41wsSecretProto\x12\x35\n\x10\x63redentials_type\x18\x01 \x01(\x0e\x32\x1b.meadowrun.Credentials.Type\x12\x13\n\x0bsecret_name\x18\x02 \x01(\t"r\n\x10\x41zureSecretProto\x12\x35\n\x10\x63redentials_type\x18\x01 \x01(\x0e\x32\x1b.meadowrun.Credentials.Type\x12\x12\n\nvault_name\x18\x02 \x01(\t\x12\x13\n\x0bsecret_name\x18\x03 \x01(\t"Z\n\x13ServerAvailableFile\x12\x35\n\x10\x63redentials_type\x18\x01 \x01(\x0e\x32\x1b.meadowrun.Credentials.Type\x12\x0c\n\x04path\x18\x02 \x01(\t"c\n\x15KubernetesSecretProto\x12\x35\n\x10\x63redentials_type\x18\x01 \x01(\x0e\x32\x1b.meadowrun.Credentials.Type\x12\x13\n\x0bsecret_name\x18\x02 \x01(\t*>\n\x0f\x45nvironmentType\x12\x0b\n\x07\x44\x45\x46\x41ULT\x10\x00\x12\t\n\x05\x43ONDA\x10\x01\x12\x07\n\x03PIP\x10\x02\x12\n\n\x06POETRY\x10\x03\x62\x06proto3'
)

_ENVIRONMENTTYPE = DESCRIPTOR.enum_types_by_name["EnvironmentType"]
EnvironmentType = enum_type_wrapper.EnumTypeWrapper(_ENVIRONMENTTYPE)
DEFAULT = 0
CONDA = 1
PIP = 2
POETRY = 3


_STRINGPAIR = DESCRIPTOR.message_types_by_name["StringPair"]
_SERVERAVAILABLEFOLDER = DESCRIPTOR.message_types_by_name["ServerAvailableFolder"]
_CODEZIPFILE = DESCRIPTOR.message_types_by_name["CodeZipFile"]
_GITREPOCOMMIT = DESCRIPTOR.message_types_by_name["GitRepoCommit"]
_GITREPOBRANCH = DESCRIPTOR.message_types_by_name["GitRepoBranch"]
_SERVERAVAILABLEINTERPRETER = DESCRIPTOR.message_types_by_name[
    "ServerAvailableInterpreter"
]
_CONTAINERATDIGEST = DESCRIPTOR.message_types_by_name["ContainerAtDigest"]
_CONTAINERATTAG = DESCRIPTOR.message_types_by_name["ContainerAtTag"]
_ENVIRONMENTSPECINCODE = DESCRIPTOR.message_types_by_name["EnvironmentSpecInCode"]
_ENVIRONMENTSPECINCODE_ADDITIONALSOFTWAREENTRY = (
    _ENVIRONMENTSPECINCODE.nested_types_by_name["AdditionalSoftwareEntry"]
)
_ENVIRONMENTSPEC = DESCRIPTOR.message_types_by_name["EnvironmentSpec"]
_ENVIRONMENTSPEC_ADDITIONALSOFTWAREENTRY = _ENVIRONMENTSPEC.nested_types_by_name[
    "AdditionalSoftwareEntry"
]
_SERVERAVAILABLECONTAINER = DESCRIPTOR.message_types_by_name["ServerAvailableContainer"]
_PYCOMMANDJOB = DESCRIPTOR.message_types_by_name["PyCommandJob"]
_QUALIFIEDFUNCTIONNAME = DESCRIPTOR.message_types_by_name["QualifiedFunctionName"]
_PYFUNCTIONJOB = DESCRIPTOR.message_types_by_name["PyFunctionJob"]
_PYAGENTJOB = DESCRIPTOR.message_types_by_name["PyAgentJob"]
_GRIDTASK = DESCRIPTOR.message_types_by_name["GridTask"]
_CONTAINERIMAGE = DESCRIPTOR.message_types_by_name["ContainerImage"]
_JOB = DESCRIPTOR.message_types_by_name["Job"]
_PROCESSSTATE = DESCRIPTOR.message_types_by_name["ProcessState"]
_JOBSTATEUPDATE = DESCRIPTOR.message_types_by_name["JobStateUpdate"]
_GRIDTASKSTATERESPONSE = DESCRIPTOR.message_types_by_name["GridTaskStateResponse"]
_CREDENTIALSSOURCEMESSAGE = DESCRIPTOR.message_types_by_name["CredentialsSourceMessage"]
_CREDENTIALS = DESCRIPTOR.message_types_by_name["Credentials"]
_AWSSECRETPROTO = DESCRIPTOR.message_types_by_name["AwsSecretProto"]
_AZURESECRETPROTO = DESCRIPTOR.message_types_by_name["AzureSecretProto"]
_SERVERAVAILABLEFILE = DESCRIPTOR.message_types_by_name["ServerAvailableFile"]
_KUBERNETESSECRETPROTO = DESCRIPTOR.message_types_by_name["KubernetesSecretProto"]
_PROCESSSTATE_PROCESSSTATEENUM = _PROCESSSTATE.enum_types_by_name["ProcessStateEnum"]
_CREDENTIALS_SERVICE = _CREDENTIALS.enum_types_by_name["Service"]
_CREDENTIALS_TYPE = _CREDENTIALS.enum_types_by_name["Type"]
StringPair = _reflection.GeneratedProtocolMessageType(
    "StringPair",
    (_message.Message,),
    {
        "DESCRIPTOR": _STRINGPAIR,
        "__module__": "meadowrun.meadowrun_pb2"
        # @@protoc_insertion_point(class_scope:meadowrun.StringPair)
    },
)
_sym_db.RegisterMessage(StringPair)

ServerAvailableFolder = _reflection.GeneratedProtocolMessageType(
    "ServerAvailableFolder",
    (_message.Message,),
    {
        "DESCRIPTOR": _SERVERAVAILABLEFOLDER,
        "__module__": "meadowrun.meadowrun_pb2"
        # @@protoc_insertion_point(class_scope:meadowrun.ServerAvailableFolder)
    },
)
_sym_db.RegisterMessage(ServerAvailableFolder)

CodeZipFile = _reflection.GeneratedProtocolMessageType(
    "CodeZipFile",
    (_message.Message,),
    {
        "DESCRIPTOR": _CODEZIPFILE,
        "__module__": "meadowrun.meadowrun_pb2"
        # @@protoc_insertion_point(class_scope:meadowrun.CodeZipFile)
    },
)
_sym_db.RegisterMessage(CodeZipFile)

GitRepoCommit = _reflection.GeneratedProtocolMessageType(
    "GitRepoCommit",
    (_message.Message,),
    {
        "DESCRIPTOR": _GITREPOCOMMIT,
        "__module__": "meadowrun.meadowrun_pb2"
        # @@protoc_insertion_point(class_scope:meadowrun.GitRepoCommit)
    },
)
_sym_db.RegisterMessage(GitRepoCommit)

GitRepoBranch = _reflection.GeneratedProtocolMessageType(
    "GitRepoBranch",
    (_message.Message,),
    {
        "DESCRIPTOR": _GITREPOBRANCH,
        "__module__": "meadowrun.meadowrun_pb2"
        # @@protoc_insertion_point(class_scope:meadowrun.GitRepoBranch)
    },
)
_sym_db.RegisterMessage(GitRepoBranch)

ServerAvailableInterpreter = _reflection.GeneratedProtocolMessageType(
    "ServerAvailableInterpreter",
    (_message.Message,),
    {
        "DESCRIPTOR": _SERVERAVAILABLEINTERPRETER,
        "__module__": "meadowrun.meadowrun_pb2"
        # @@protoc_insertion_point(class_scope:meadowrun.ServerAvailableInterpreter)
    },
)
_sym_db.RegisterMessage(ServerAvailableInterpreter)

ContainerAtDigest = _reflection.GeneratedProtocolMessageType(
    "ContainerAtDigest",
    (_message.Message,),
    {
        "DESCRIPTOR": _CONTAINERATDIGEST,
        "__module__": "meadowrun.meadowrun_pb2"
        # @@protoc_insertion_point(class_scope:meadowrun.ContainerAtDigest)
    },
)
_sym_db.RegisterMessage(ContainerAtDigest)

ContainerAtTag = _reflection.GeneratedProtocolMessageType(
    "ContainerAtTag",
    (_message.Message,),
    {
        "DESCRIPTOR": _CONTAINERATTAG,
        "__module__": "meadowrun.meadowrun_pb2"
        # @@protoc_insertion_point(class_scope:meadowrun.ContainerAtTag)
    },
)
_sym_db.RegisterMessage(ContainerAtTag)

EnvironmentSpecInCode = _reflection.GeneratedProtocolMessageType(
    "EnvironmentSpecInCode",
    (_message.Message,),
    {
        "AdditionalSoftwareEntry": _reflection.GeneratedProtocolMessageType(
            "AdditionalSoftwareEntry",
            (_message.Message,),
            {
                "DESCRIPTOR": _ENVIRONMENTSPECINCODE_ADDITIONALSOFTWAREENTRY,
                "__module__": "meadowrun.meadowrun_pb2"
                # @@protoc_insertion_point(class_scope:meadowrun.EnvironmentSpecInCode.AdditionalSoftwareEntry)
            },
        ),
        "DESCRIPTOR": _ENVIRONMENTSPECINCODE,
        "__module__": "meadowrun.meadowrun_pb2"
        # @@protoc_insertion_point(class_scope:meadowrun.EnvironmentSpecInCode)
    },
)
_sym_db.RegisterMessage(EnvironmentSpecInCode)
_sym_db.RegisterMessage(EnvironmentSpecInCode.AdditionalSoftwareEntry)

EnvironmentSpec = _reflection.GeneratedProtocolMessageType(
    "EnvironmentSpec",
    (_message.Message,),
    {
        "AdditionalSoftwareEntry": _reflection.GeneratedProtocolMessageType(
            "AdditionalSoftwareEntry",
            (_message.Message,),
            {
                "DESCRIPTOR": _ENVIRONMENTSPEC_ADDITIONALSOFTWAREENTRY,
                "__module__": "meadowrun.meadowrun_pb2"
                # @@protoc_insertion_point(class_scope:meadowrun.EnvironmentSpec.AdditionalSoftwareEntry)
            },
        ),
        "DESCRIPTOR": _ENVIRONMENTSPEC,
        "__module__": "meadowrun.meadowrun_pb2"
        # @@protoc_insertion_point(class_scope:meadowrun.EnvironmentSpec)
    },
)
_sym_db.RegisterMessage(EnvironmentSpec)
_sym_db.RegisterMessage(EnvironmentSpec.AdditionalSoftwareEntry)

ServerAvailableContainer = _reflection.GeneratedProtocolMessageType(
    "ServerAvailableContainer",
    (_message.Message,),
    {
        "DESCRIPTOR": _SERVERAVAILABLECONTAINER,
        "__module__": "meadowrun.meadowrun_pb2"
        # @@protoc_insertion_point(class_scope:meadowrun.ServerAvailableContainer)
    },
)
_sym_db.RegisterMessage(ServerAvailableContainer)

PyCommandJob = _reflection.GeneratedProtocolMessageType(
    "PyCommandJob",
    (_message.Message,),
    {
        "DESCRIPTOR": _PYCOMMANDJOB,
        "__module__": "meadowrun.meadowrun_pb2"
        # @@protoc_insertion_point(class_scope:meadowrun.PyCommandJob)
    },
)
_sym_db.RegisterMessage(PyCommandJob)

QualifiedFunctionName = _reflection.GeneratedProtocolMessageType(
    "QualifiedFunctionName",
    (_message.Message,),
    {
        "DESCRIPTOR": _QUALIFIEDFUNCTIONNAME,
        "__module__": "meadowrun.meadowrun_pb2"
        # @@protoc_insertion_point(class_scope:meadowrun.QualifiedFunctionName)
    },
)
_sym_db.RegisterMessage(QualifiedFunctionName)

PyFunctionJob = _reflection.GeneratedProtocolMessageType(
    "PyFunctionJob",
    (_message.Message,),
    {
        "DESCRIPTOR": _PYFUNCTIONJOB,
        "__module__": "meadowrun.meadowrun_pb2"
        # @@protoc_insertion_point(class_scope:meadowrun.PyFunctionJob)
    },
)
_sym_db.RegisterMessage(PyFunctionJob)

PyAgentJob = _reflection.GeneratedProtocolMessageType(
    "PyAgentJob",
    (_message.Message,),
    {
        "DESCRIPTOR": _PYAGENTJOB,
        "__module__": "meadowrun.meadowrun_pb2"
        # @@protoc_insertion_point(class_scope:meadowrun.PyAgentJob)
    },
)
_sym_db.RegisterMessage(PyAgentJob)

GridTask = _reflection.GeneratedProtocolMessageType(
    "GridTask",
    (_message.Message,),
    {
        "DESCRIPTOR": _GRIDTASK,
        "__module__": "meadowrun.meadowrun_pb2"
        # @@protoc_insertion_point(class_scope:meadowrun.GridTask)
    },
)
_sym_db.RegisterMessage(GridTask)

ContainerImage = _reflection.GeneratedProtocolMessageType(
    "ContainerImage",
    (_message.Message,),
    {
        "DESCRIPTOR": _CONTAINERIMAGE,
        "__module__": "meadowrun.meadowrun_pb2"
        # @@protoc_insertion_point(class_scope:meadowrun.ContainerImage)
    },
)
_sym_db.RegisterMessage(ContainerImage)

Job = _reflection.GeneratedProtocolMessageType(
    "Job",
    (_message.Message,),
    {
        "DESCRIPTOR": _JOB,
        "__module__": "meadowrun.meadowrun_pb2"
        # @@protoc_insertion_point(class_scope:meadowrun.Job)
    },
)
_sym_db.RegisterMessage(Job)

ProcessState = _reflection.GeneratedProtocolMessageType(
    "ProcessState",
    (_message.Message,),
    {
        "DESCRIPTOR": _PROCESSSTATE,
        "__module__": "meadowrun.meadowrun_pb2"
        # @@protoc_insertion_point(class_scope:meadowrun.ProcessState)
    },
)
_sym_db.RegisterMessage(ProcessState)

JobStateUpdate = _reflection.GeneratedProtocolMessageType(
    "JobStateUpdate",
    (_message.Message,),
    {
        "DESCRIPTOR": _JOBSTATEUPDATE,
        "__module__": "meadowrun.meadowrun_pb2"
        # @@protoc_insertion_point(class_scope:meadowrun.JobStateUpdate)
    },
)
_sym_db.RegisterMessage(JobStateUpdate)

GridTaskStateResponse = _reflection.GeneratedProtocolMessageType(
    "GridTaskStateResponse",
    (_message.Message,),
    {
        "DESCRIPTOR": _GRIDTASKSTATERESPONSE,
        "__module__": "meadowrun.meadowrun_pb2"
        # @@protoc_insertion_point(class_scope:meadowrun.GridTaskStateResponse)
    },
)
_sym_db.RegisterMessage(GridTaskStateResponse)

CredentialsSourceMessage = _reflection.GeneratedProtocolMessageType(
    "CredentialsSourceMessage",
    (_message.Message,),
    {
        "DESCRIPTOR": _CREDENTIALSSOURCEMESSAGE,
        "__module__": "meadowrun.meadowrun_pb2"
        # @@protoc_insertion_point(class_scope:meadowrun.CredentialsSourceMessage)
    },
)
_sym_db.RegisterMessage(CredentialsSourceMessage)

Credentials = _reflection.GeneratedProtocolMessageType(
    "Credentials",
    (_message.Message,),
    {
        "DESCRIPTOR": _CREDENTIALS,
        "__module__": "meadowrun.meadowrun_pb2"
        # @@protoc_insertion_point(class_scope:meadowrun.Credentials)
    },
)
_sym_db.RegisterMessage(Credentials)

AwsSecretProto = _reflection.GeneratedProtocolMessageType(
    "AwsSecretProto",
    (_message.Message,),
    {
        "DESCRIPTOR": _AWSSECRETPROTO,
        "__module__": "meadowrun.meadowrun_pb2"
        # @@protoc_insertion_point(class_scope:meadowrun.AwsSecretProto)
    },
)
_sym_db.RegisterMessage(AwsSecretProto)

AzureSecretProto = _reflection.GeneratedProtocolMessageType(
    "AzureSecretProto",
    (_message.Message,),
    {
        "DESCRIPTOR": _AZURESECRETPROTO,
        "__module__": "meadowrun.meadowrun_pb2"
        # @@protoc_insertion_point(class_scope:meadowrun.AzureSecretProto)
    },
)
_sym_db.RegisterMessage(AzureSecretProto)

ServerAvailableFile = _reflection.GeneratedProtocolMessageType(
    "ServerAvailableFile",
    (_message.Message,),
    {
        "DESCRIPTOR": _SERVERAVAILABLEFILE,
        "__module__": "meadowrun.meadowrun_pb2"
        # @@protoc_insertion_point(class_scope:meadowrun.ServerAvailableFile)
    },
)
_sym_db.RegisterMessage(ServerAvailableFile)

KubernetesSecretProto = _reflection.GeneratedProtocolMessageType(
    "KubernetesSecretProto",
    (_message.Message,),
    {
        "DESCRIPTOR": _KUBERNETESSECRETPROTO,
        "__module__": "meadowrun.meadowrun_pb2"
        # @@protoc_insertion_point(class_scope:meadowrun.KubernetesSecretProto)
    },
)
_sym_db.RegisterMessage(KubernetesSecretProto)

if _descriptor._USE_C_DESCRIPTORS == False:

    DESCRIPTOR._options = None
    _ENVIRONMENTSPECINCODE_ADDITIONALSOFTWAREENTRY._options = None
    _ENVIRONMENTSPECINCODE_ADDITIONALSOFTWAREENTRY._serialized_options = b"8\001"
    _ENVIRONMENTSPEC_ADDITIONALSOFTWAREENTRY._options = None
    _ENVIRONMENTSPEC_ADDITIONALSOFTWAREENTRY._serialized_options = b"8\001"
    _ENVIRONMENTTYPE._serialized_start = 4591
    _ENVIRONMENTTYPE._serialized_end = 4653
    _STRINGPAIR._serialized_start = 40
    _STRINGPAIR._serialized_end = 80
    _SERVERAVAILABLEFOLDER._serialized_start = 82
    _SERVERAVAILABLEFOLDER._serialized_end = 125
    _CODEZIPFILE._serialized_start = 127
    _CODEZIPFILE._serialized_end = 191
    _GITREPOCOMMIT._serialized_start = 193
    _GITREPOCOMMIT._serialized_end = 266
    _GITREPOBRANCH._serialized_start = 268
    _GITREPOBRANCH._serialized_end = 341
    _SERVERAVAILABLEINTERPRETER._serialized_start = 343
    _SERVERAVAILABLEINTERPRETER._serialized_end = 397
    _CONTAINERATDIGEST._serialized_start = 399
    _CONTAINERATDIGEST._serialized_end = 454
    _CONTAINERATTAG._serialized_start = 456
    _CONTAINERATTAG._serialized_end = 505
    _ENVIRONMENTSPECINCODE._serialized_start = 508
    _ENVIRONMENTSPECINCODE._serialized_end = 777
    _ENVIRONMENTSPECINCODE_ADDITIONALSOFTWAREENTRY._serialized_start = 720
    _ENVIRONMENTSPECINCODE_ADDITIONALSOFTWAREENTRY._serialized_end = 777
    _ENVIRONMENTSPEC._serialized_start = 780
    _ENVIRONMENTSPEC._serialized_end = 1048
    _ENVIRONMENTSPEC_ADDITIONALSOFTWAREENTRY._serialized_start = 720
    _ENVIRONMENTSPEC_ADDITIONALSOFTWAREENTRY._serialized_end = 777
    _SERVERAVAILABLECONTAINER._serialized_start = 1050
    _SERVERAVAILABLECONTAINER._serialized_end = 1096
    _PYCOMMANDJOB._serialized_start = 1098
    _PYCOMMANDJOB._serialized_end = 1169
    _QUALIFIEDFUNCTIONNAME._serialized_start = 1171
    _QUALIFIEDFUNCTIONNAME._serialized_end = 1238
    _PYFUNCTIONJOB._serialized_start = 1241
    _PYFUNCTIONJOB._serialized_end = 1406
    _PYAGENTJOB._serialized_start = 1409
    _PYAGENTJOB._serialized_end = 1609
    _GRIDTASK._serialized_start = 1611
    _GRIDTASK._serialized_end = 1691
    _CONTAINERIMAGE._serialized_start = 1694
    _CONTAINERIMAGE._serialized_end = 1938
    _JOB._serialized_start = 1941
    _JOB._serialized_end = 3053
    _PROCESSSTATE._serialized_start = 3056
    _PROCESSSTATE._serialized_end = 3506
    _PROCESSSTATE_PROCESSSTATEENUM._serialized_start = 3261
    _PROCESSSTATE_PROCESSSTATEENUM._serialized_end = 3506
    _JOBSTATEUPDATE._serialized_start = 3508
    _JOBSTATEUPDATE._serialized_end = 3588
    _GRIDTASKSTATERESPONSE._serialized_start = 3590
    _GRIDTASKSTATERESPONSE._serialized_end = 3695
    _CREDENTIALSSOURCEMESSAGE._serialized_start = 3698
    _CREDENTIALSSOURCEMESSAGE._serialized_end = 4034
    _CREDENTIALS._serialized_start = 4037
    _CREDENTIALS._serialized_end = 4186
    _CREDENTIALS_SERVICE._serialized_start = 4073
    _CREDENTIALS_SERVICE._serialized_end = 4124
    _CREDENTIALS_TYPE._serialized_start = 4126
    _CREDENTIALS_TYPE._serialized_end = 4186
    _AWSSECRETPROTO._serialized_start = 4188
    _AWSSECRETPROTO._serialized_end = 4280
    _AZURESECRETPROTO._serialized_start = 4282
    _AZURESECRETPROTO._serialized_end = 4396
    _SERVERAVAILABLEFILE._serialized_start = 4398
    _SERVERAVAILABLEFILE._serialized_end = 4488
    _KUBERNETESSECRETPROTO._serialized_start = 4490
    _KUBERNETESSECRETPROTO._serialized_end = 4589
# @@protoc_insertion_point(module_scope)
